import os

import numpy as np
import scipy.constants as ct
from pytest import approx, raises

from wake_t.utilities.bunch_generation import get_gaussian_bunch_from_twiss
from wake_t.fields.analytical_field import AnalyticalField
from wake_t.beamline_elements import FieldElement
from wake_t.diagnostics import analyze_bunch


tests_output_folder = "./tests_output"


def b_x(x, y, z, t, bx, constants):
    """B_x component."""
    k = constants[0]
    bx -= k * y


def b_y(x, y, z, t, by, constants):
    """B_y component."""
    k = constants[0]
    by += k * x


def test_field_element_tracking():
    """Test that tracking in a field element works as expected by tracking
    a bunch through an analytical field (azimuthal magnetic field)"""
    output_folder = os.path.join(tests_output_folder, "test_field_element_tracking")

    # Set numpy random seed to get reproducible results
    np.random.seed(1)

    # Define field.
    b_theta = AnalyticalField(b_x=b_x, b_y=b_y, constants=[-100])

    # Create bunch.
    emitt_nx = emitt_ny = 1e-6  # m
    beta_x = beta_y = 1.0  # m
    s_t = 100.0  # fs
    gamma_avg = 1000
    ene_spread = 0.1  # %
    q_bunch = 30  # pC
    xi_avg = 0.0  # m
    n_part = 1e4
    bunch = get_gaussian_bunch_from_twiss(
        en_x=emitt_nx,
        en_y=emitt_ny,
        a_x=0,
        a_y=0,
        b_x=beta_x,
        b_y=beta_y,
        ene=gamma_avg,
        ene_sp=ene_spread,
        s_t=s_t,
        xi_c=xi_avg,
        q_tot=q_bunch,
        n_part=n_part,
        name="elec_bunch",
    )

    # Create field element.
    element = FieldElement(length=1, dt_bunch=1e-3 / ct.c, n_out=10, fields=[b_theta])

    # Do tracking.
    diag_dir = os.path.join(output_folder, "diags")
    element.track(bunch, opmd_diag=True, diag_dir=diag_dir)

    # Check that results have not changed.
    bunch_params = analyze_bunch(bunch)
    beta_x = bunch_params["beta_x"]
    assert approx(beta_x, rel=1e-10) == 0.05450126902819044


def test_field_element_error():
    """Test that the expected errors are raised"""

    # Create bunch.
    emitt_nx = emitt_ny = 1e-6  # m
    beta_x = beta_y = 1.0  # m
    s_t = 100.0  # fs
    gamma_avg = 1000
    ene_spread = 0.1  # %
    q_bunch = 30  # pC
    xi_avg = 0.0  # m
    n_part = 1e4
    bunch = get_gaussian_bunch_from_twiss(
        en_x=emitt_nx,
        en_y=emitt_ny,
        a_x=0,
        a_y=0,
        b_x=beta_x,
        b_y=beta_y,
        ene=gamma_avg,
        ene_sp=ene_spread,
        s_t=s_t,
        xi_c=xi_avg,
        q_tot=q_bunch,
        n_part=n_part,
        name="elec_bunch",
    )

    # Create field element with too many dt_bunch.
    element = FieldElement(
        length=1,
        dt_bunch=[1e-3 / ct.c, 1e-3 / ct.c],
        n_out=10,
        fields=[],
    )

    # Check that an error is raised because the number of `dt_bunch` does
    # not agree with the number of bunches.
    with raises(ValueError):
        element.track(bunch, opmd_diag=False)
    # This one should instead work.
    element.track([bunch, bunch.copy()], opmd_diag=False)


if __name__ == "__main__":
    test_field_element_tracking()
    test_field_element_error()



import numpy as np
import matplotlib.pyplot as plt
import scipy.constants as ct
import aptools.plasma_accel.general_equations as ge
from wake_t.utilities.bunch_generation import get_gaussian_bunch_from_size
from wake_t.physics_models.plasma_wakefields.qs_rz_baxevanis.solver import (
    calculate_beam_source,
)


def test_sc_gaussian_beam(show=False):
    """Test the space-charge calculation used in the `quasistatic_2d`
    wakefield model.

    A Gaussian beam is generated and the calculated space-charge field is
    compared against the analytical expectation.

    Parameters
    ----------
    show : bool, optional
        Whether to show a lineout of the fields, by default False
    """
    # Set seed fo reproducible results.
    np.random.seed(0)

    # Plasma density.
    n_p = 1e23
    s_d = ge.plasma_skin_depth(n_p * 1e-6)
    E0 = ge.plasma_cold_non_relativisct_wave_breaking_field(n_p * 1e-6)

    # Bunch parameters
    s_x = 1e-6
    en = 1e-6
    gamma = 200
    ene_sp = 1e-5
    s_t = 10
    q_tot = 100
    n_part = 1e7
    s_z = s_t * (1e-15 * ct.c)
    s_r = s_x

    # Generate bunch
    bunch = get_gaussian_bunch_from_size(
        en, en, s_x, s_x, gamma, ene_sp, s_t, 0, q_tot, n_part=n_part
    )

    # Resolutions ans shapes to test
    n_r_test = [20, 100]
    n_z_test = [41, 101]
    p_shape_test = ["linear", "cubic"]

    for n_r, n_z in zip(n_r_test, n_z_test):
        for p_shape in p_shape_test:
            # Generate grid
            z_min = -10e-6 / s_d
            z_max = 10e-6 / s_d
            r_max = 10e-6 / s_d
            dr = r_max / n_r
            dz = (z_max - z_min) / (n_z - 1)
            r_fld = np.linspace(dr / 2, r_max - dr / 2, n_r)
            z_fld = np.linspace(z_min, z_max, n_z)

            b_theta_grid = np.zeros((n_z + 4, n_r + 4))
            calculate_beam_source(
                bunch, n_p, n_r, n_z, r_fld[0], z_fld[0], dr, dz, p_shape, b_theta_grid
            )

            # Remove guard cells.
            b_theta_grid = b_theta_grid[2:-2, 2:-2]

            # Convert to SI.
            b_theta_grid *= E0 / ct.c

            # Deposited density at beam center.
            i_center = int((n_z - 1) / 2)

            # Calculate azimuthal magnetic field (analytical and from grid)
            b_theta_ana_0 = b_theta_gaussian_beam(
                r_fld * s_d, 0.0, q_tot, s_r, s_z, gamma
            )
            b_theta_grid_0 = b_theta_grid[i_center]

            # Relative deviation w.r.t. analytical.
            rel_b_dev = np.abs((b_theta_grid_0 - b_theta_ana_0) / b_theta_ana_0)

            # Check long range fields are accurate.
            assert rel_b_dev[-1] < 3e-3

            # Check reasonable deviation in general. At low res, ~10%
            # differences can occur close to axis.
            assert np.max(rel_b_dev) < 13e-2

            if show:
                plt.plot(r_fld, b_theta_ana_0)
                plt.plot(r_fld, b_theta_grid_0)

                plt.show()


def b_theta_gaussian_beam(r, z, q_tot, s_r, s_z, gamma):
    """Analytical azimuthal magnetic field for a Gaussian beam."""
    beta = np.sqrt(1 - 1 / gamma**2)
    e_r = (
        1
        / ((2 * np.pi) ** (3 / 2) * ct.epsilon_0)
        * (-q_tot)
        * 1e-12
        / s_z
        * np.exp(-(z**2) / (2 * s_z**2))
        * (1 - np.exp(-(r**2) / (2 * s_r**2)))
        / r
    )
    b_theta = beta / ct.c * e_r
    return b_theta


if __name__ == "__main__":
    test_sc_gaussian_beam()



"""This module contains the Tracker class."""

from typing import Optional, Callable, List, Literal

import numpy as np
import scipy.constants as ct

from wake_t.particles.particle_bunch import ParticleBunch
from wake_t.fields.base import Field
from wake_t.fields.analytical_field import AnalyticalField
from wake_t.fields.numerical_field import NumericalField
from wake_t.diagnostics.openpmd_diag import OpenPMDDiagnostics
from wake_t.utilities.numba import num_threads, set_num_threads, get_num_threads
from .progress_bar import get_progress_bar


class Tracker:
    """Class in charge of evolving in time the particle bunches and fields.

    There are 3 main ingredients in the simulation of an accelerator stage:
    particle bunches, electromagnetic fields, and diagnostics. Each of these
    elements has a certain periodicity: the bunches are evolved with a certain
    time step (which can be adaptive), the numerical fields are also updated
    with a certain time step, and the diagnostics are generated with a regular
    periodicity.

    Thus, in time, a simulation looks something like:

        Bunch: x-x-x--x---x-----x------x--------x---------x----x|
        Field: x-------x-------x-------x-------x-------x-------x|
        Diags: x-----------x-----------x-----------x-----------x|
        Time:  ------------------------------------------------>| End

    Where the `x` denote the moments in time where each quantity is updated.

    The job of the `Tracker` is to orchestrate this flow and update each
    element at the right time.

    Parameters
    ----------
    t_final : float
        Final time of the tracking.
    bunches : list, optional
        List of `ParticleBunch`es to track.
    dt_bunches : list, optional
        List of time steps. There should be one value per bunch. Possible
        values are any float >0 for a constant time step, or 'auto' for
        enabling an adaptive time step.
    fields : list, optional
        List of `Field`s in which to evolve the particles.
    n_diags : int, optional
        Number of diagnostics to output.
    opmd_diags : OpenPMDDiagnostics, optional
        Instace of OpenPMDDiagnostics in charge of generating the openPMD
        output. If not given, no diagnostics will be generated.
    auto_dt_bunch_f : callable, optional
        Function used to determine the adaptive time step for bunches in
        which the time step is set to `'auto'`. The function should take
        solely a `ParticleBunch` as argument.
    bunch_pusher : str, optional
        The particle pusher used to evolve the bunches. Possible values
        are `'boris'` or `'rk4'`.
    push_bunches_before_diags : bool, optional
        Whether to push the bunches before saving them to the diagnostics.
        Since the time step of the diagnostics can be different from that
        of the bunches, it could happen that the bunches appear in the
        diagnostics as they were at the last push, but not at the actual
        time of the diagnostics. Setting this parameter to ``True``
        (default) ensures that an additional push is given to all bunches
        to evolve them to the diagnostics time before saving.
        This additional push will always have a time step smaller than
        the the time step of the bunch, so it has no detrimental impact
        on the accuracy of the simulation. However, it could make
        convergence studies more difficult to interpret,
        since the number of pushes will depend on `n_diags`. Therefore,
        it is exposed as an option so that it can be disabled if needed.
    show_progress_bar : bool, optional
        Whether to show a progress bar of the tracking. By default ``True``.
    section_name : str, optional
        Name of the section to be tracked. This will be appended to the
        beginning of the progress bar.

    """

    def __init__(
        self,
        t_final: float,
        bunches: Optional[List[ParticleBunch]] = [],
        dt_bunches: Optional[List[float]] = [],
        fields: Optional[List[Field]] = [],
        n_diags: Optional[int] = 0,
        opmd_diags: Optional[OpenPMDDiagnostics] = None,
        auto_dt_bunch_f: Optional[Callable[[ParticleBunch], float]] = None,
        bunch_pusher: Optional[Literal["boris", "rk4"]] = "boris",
        push_bunches_before_diags: Optional[bool] = True,
        show_progress_bar: Optional[bool] = True,
        section_name: Optional[str] = "Simulation",
    ) -> None:
        self.t_final = t_final
        self.bunches = bunches
        self.dt_bunches = dt_bunches
        self.fields = fields if len(fields) > 0 else [AnalyticalField()]
        self.opmd_diags = opmd_diags
        self.n_diags = n_diags
        self.auto_dt_bunch_f = auto_dt_bunch_f
        self.bunch_pusher = bunch_pusher
        self.push_bunches_before_diags = push_bunches_before_diags
        self.show_progress_bar = show_progress_bar
        self.section_name = section_name

        # Get all numerical fields and their time steps.
        self.num_fields = [f for f in fields if isinstance(f, NumericalField)]
        for field in self.num_fields:
            field.adjust_dt(self.t_final)
        self.dt_fields = [f.dt_update for f in self.num_fields]

        # Make lists with all objects to track and their time steps.
        self.objects_to_track = [*self.bunches, *self.num_fields]
        self.dt_objects = [*self.dt_bunches, *self.dt_fields]

        # Get list of bunches with adaptive time step.
        self.auto_dt_bunches = []
        for i, dt in enumerate(self.dt_bunches):
            if dt == "auto":
                self.auto_dt_bunches.append(self.bunches[i])

        # If needed, add diagnostics to objects to track.
        if self.n_diags > 0:
            self.objects_to_track.append("diags")
            self.dt_diags = self.t_final / self.n_diags
            self.dt_objects.append(self.dt_diags)
            self.bunch_list = []
            for bunch in bunches:
                self.bunch_list.append([])

        # Initialize tracking time.
        self.t_tracking = 0.0

    def do_tracking(self) -> List[List[ParticleBunch]]:
        """Do the tracking.

        Returns
        -------
        list
            A list with `n` items, where `n` is the number of bunches to track.
            Each item is another list with `n_diag` copies of the particle
            bunch along the tracking.
        """
        # Get current number of threads before setting it to the number
        # requested by Wake-T.
        num_threads_outside_waket = get_num_threads()
        set_num_threads(num_threads)

        # Initialize progress bar.
        progress_bar = get_progress_bar(
            description=self.section_name,
            total_length=self.t_final * ct.c,
            disable=not self.show_progress_bar,
        )

        # Calculate fields at t=0.
        for field in self.num_fields:
            field.update(self.bunches)

        # Generate initial diagnostics.
        self.generate_diagnostics()

        # Allocate arrays containing the time step and current time of all
        # objects during tracking.
        t_objects = np.zeros(len(self.dt_objects))
        dt_objects = np.zeros(len(self.dt_objects))

        # Fill up time steps.
        for i, dt in enumerate(self.dt_objects):
            if dt == "auto":
                dt_objects[i] = self.auto_dt_bunch_f(self.bunches[i])
            else:
                dt_objects[i] = dt

        # Start tracking loop.
        while True:
            # Calculate next time of all objects.
            t_next_objects = t_objects + dt_objects

            # Get index of the object with smallest next time.
            # float32 is used to avoid precision issues. Since
            # `t_next_objects` is calculated by repeatedly adding dt_objects,
            # it can happen that two objects that are supposed to have exactly
            # the same `t_next_objects` do not have it due to precision issues.
            # Going from float64 to float32 reduces the number of decimals,
            # thus making sure that two "almost identical" numbers at float64
            # are actually identical as float32.
            i_next = np.argmin(t_next_objects.astype(np.float32))

            # Get next object and its corresponding time and time step.
            obj_next = self.objects_to_track[i_next]
            dt_next = dt_objects[i_next]
            t_next = t_next_objects[i_next]
            t_current = t_objects[i_next]

            # If the next time of the object is beyond `t_final`, the tracking
            # is finished.
            if np.float32(t_next) > np.float32(self.t_final):
                break

            # Advance tracking time.
            self.t_tracking = t_next

            # If next object is a ParticleBunch, update it.
            if isinstance(obj_next, ParticleBunch):
                self.evolve_bunch(
                    bunch=obj_next,
                    t_current=t_current,
                    t_next=t_next,
                    dt_next=dt_next,
                    i_next=i_next,
                    dt_objects=dt_objects,
                )

            # If next object is a NumericalField, update it.
            elif isinstance(obj_next, NumericalField):
                obj_next.update(self.bunches)

            # If next object are the diagnostics, generate them.
            elif obj_next == "diags":
                # Evolve all bunches to the diagnostics time.
                if self.push_bunches_before_diags:
                    for i, obj in enumerate(self.objects_to_track):
                        if isinstance(obj, ParticleBunch):
                            dt_bunch = t_next - t_objects[i]
                            self.evolve_bunch(
                                bunch=obj,
                                t_current=t_objects[i],
                                t_next=t_next,
                                dt_next=dt_bunch,
                                i_next=i,
                                dt_objects=dt_objects,
                            )
                            t_objects[i] += dt_bunch
                self.generate_diagnostics()

            # Advance current time of the update object.
            t_objects[i_next] += dt_next

            # Update progress bar.
            progress_bar.update(self.t_tracking * ct.c - progress_bar.n)

        # Finalize tracking by increasing z position of diagnostics.
        if self.opmd_diags is not None:
            self.opmd_diags.increase_z_pos(self.t_final * ct.c)

        # Close progress bar.
        progress_bar.close()

        # Reset the number of threads to the value outside of Wake-T.
        # This should help avoiding Wake-T to affect the behavior of other
        # applications that require a different number of threads.
        set_num_threads(num_threads_outside_waket)

        return self.bunch_list

    def evolve_bunch(
        self,
        bunch: ParticleBunch,
        t_current: float,
        t_next: float,
        dt_next: float,
        i_next: int,
        dt_objects: List,
    ):
        """Evolve particle bunch to next time step.

        Parameters
        ----------
        bunch : ParticleBunch
            The particle bunch to evolve.
        t_current : float
            The current time of the simulation.
        t_next : float
            The time of the next step.
        dt_next : float
            The time step by which to advance the bunch.
        i_next : int
            The index of the bunch in the list of objects to track.
        dt_objects : List
            The time steps of all objects to track.
        """
        bunch.evolve(self.fields, t_current, dt_next, self.bunch_pusher)
        # Update the time step if set to `'auto'`.
        if bunch in self.auto_dt_bunches:
            dt_objects[i_next] = self.auto_dt_bunch_f(bunch)
        # Determine if this was the last push.
        final_push = np.float32(t_next) == np.float32(self.t_final)
        # Determine if next push brings the bunch beyond `t_final`.
        next_push_beyond_final_time = t_next + dt_objects[i_next] > self.t_final
        # Make sure the last push of the bunch advances it to exactly
        # `t_final`.
        if not final_push and next_push_beyond_final_time:
            dt_objects[i_next] = self.t_final - t_next

    def generate_diagnostics(self) -> None:
        """Generate tracking diagnostics."""
        # Make copy of current bunches and store in output list.
        for i, bunch in enumerate(self.bunches):
            self.bunch_list[i].append(bunch.copy())

        # If needed, write also the openPMD diagnostics.
        if self.opmd_diags is not None:
            self.opmd_diags.write_diagnostics(
                self.t_tracking, self.dt_diags, self.bunches, self.fields
            )



from .openpmd_diag import OpenPMDDiagnostics
from .bunch_analysis import (
    analyze_bunch,
    analyze_bunch_list,
    save_parameters_to_file,
    read_parameters_from_file,
)


__all__ = [
    "OpenPMDDiagnostics",
    "analyze_bunch",
    "analyze_bunch_list",
    "save_parameters_to_file",
    "read_parameters_from_file",
]



from typing import Optional, Union, List

from wake_t.diagnostics import OpenPMDDiagnostics
from wake_t.particles.particle_bunch import ParticleBunch


class Beamline:
    """
    Class for grouping beamline elements and allowing easier tracking.

    """

    def __init__(self, elements: List) -> None:
        self.elements = elements

    def track(
        self,
        bunches: Optional[Union[ParticleBunch, List[ParticleBunch]]] = [],
        opmd_diag: Optional[bool] = False,
        diag_dir: Optional[str] = None,
        show_progress_bar: Optional[bool] = True,
    ) -> Union[List[ParticleBunch], List[List[ParticleBunch]]]:
        """
        Track bunch through beamline.

        Parameters
        ----------
        bunches : ParticleBunch or list of ParticleBunch
            Particle bunches to be tracked.
        opmd_diag : bool or OpenPMDDiagnostics
            Determines whether to write simulation diagnostics to disk (i.e.
            particle distributions and fields). The output is written to
            HDF5 files following the openPMD standard. The number of outputs
            per beamline element is determined by its `n_out` value. It is also
            possible to provide an already existing OpenPMDDiagnostics
            instance instead of a boolean value.
        diag_dir : str
            Directory into which the openPMD output will be written. By default
            this is a 'diags' folder in the current directory. Only needed if
            `opmd_diag=True`.
        show_progress_bar : bool, optional
            Whether to show a progress bar of the tracking through each
            element. By default ``True``.

        Returns
        -------
        A list of size 'n_out' containing the bunch distribution at each step.

        """
        bunch_list = []
        if type(opmd_diag) is not OpenPMDDiagnostics and opmd_diag:
            opmd_diag = OpenPMDDiagnostics(write_dir=diag_dir)
        for element in self.elements:
            bunch_list.extend(
                element.track(
                    bunches,
                    opmd_diag=opmd_diag,
                    show_progress_bar=show_progress_bar,
                )
            )
        return bunch_list



from pytest import approx
import matplotlib.pyplot as plt
import numpy as np
from aptools.plotting.quick_diagnostics import slice_analysis

from wake_t import PlasmaStage, GaussianPulse
from wake_t.utilities.bunch_generation import get_matched_bunch
from wake_t.diagnostics import analyze_bunch_list


def test_fluid_model(plot=False):
    """Test the 1D fluid wakefield model."""
    # Set numpy random seed to get reproducible results.
    np.random.seed(0)

    # Create laser driver.
    laser = GaussianPulse(100e-6, l_0=800e-9, w_0=70e-6, a_0=0.8, tau=30e-15, z_foc=0.0)

    # Create bunch (matched to a focusing strength of 0.13 MT/m).
    en = 1e-6  # m
    ene = 200  # units of beta*gamma
    ene_sp = 0.3  # %
    xi_c = laser.xi_c - 55e-6  # m
    s_t = 3  # fs
    q_tot = 1  # pC
    n_part = 1e4
    bunch = get_matched_bunch(en, en, ene, ene_sp, s_t, xi_c, q_tot, n_part, k_x=0.13e6)

    # Create plasma stage.
    plasma = PlasmaStage(
        1e-2,
        1e23,
        laser=laser,
        wakefield_model="cold_fluid_1d",
        n_out=50,
        laser_evolution=True,
        beam_wakefields=True,
        dz_fields=0.5e-3,
        r_max=200e-6,
        xi_min=40e-6,
        xi_max=120e-6,
        n_r=200,
        n_xi=200,
    )

    # Do tracking.
    bunch_list = plasma.track(bunch)

    # Analyze bunch evolution.
    params_evolution = analyze_bunch_list(bunch_list)

    # Check final parameters.
    ene_sp = params_evolution["rel_ene_spread"][-1]
    assert approx(ene_sp, rel=1e-10) == 0.024157374662580212

    # Quick plot of results.
    if plot:
        z = params_evolution["prop_dist"] * 1e2
        plt.figure()
        plt.subplot(411)
        plt.plot(z, params_evolution["beta_x"] * 1e3)
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\beta_x$ [mm]")
        plt.subplot(412)
        plt.plot(z, params_evolution["emitt_x"] * 1e6)
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\epsilon_{nx}$ [$\\mu$m]")
        plt.subplot(413)
        plt.plot(z, params_evolution["rel_ene_spread"] * 100)
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\frac{\\Delta \\gamma}{\\gamma}$ [%]")
        plt.subplot(414)
        plt.plot(z, params_evolution["avg_ene"])
        plt.xlabel("z [cm]")
        plt.ylabel("$\\gamma$")
        plt.tight_layout()
        fig_2 = plt.figure()
        slice_analysis(
            bunch.x, bunch.y, bunch.xi, bunch.px, bunch.py, bunch.pz, bunch.q, fig=fig_2
        )
        plt.show()


if __name__ == "__main__":
    test_fluid_model(plot=True)



import numpy as np

from wake_t.utilities.numba import njit_serial
from wake_t.particles.interpolation import gather_sources_qs_baxevanis
from ..psi_and_derivatives import calculate_psi_and_derivatives_at_particles
from ..b_theta import calculate_b_theta_at_particles


@njit_serial()
def evolve_plasma_rk4(
    dxi,
    dr,
    xi,
    r,
    pr,
    gamma,
    q,
    r_max_plasma,
    dr_p,
    pc,
    a2,
    nabla_a2,
    b_t_0,
    r_fld,
    xi_fld,
    dr_1,
    dr_2,
    dr_3,
    dr_4,
    dpr_1,
    dpr_2,
    dpr_3,
    dpr_4,
    a2_1,
    nabla_a2_1,
    b_t_0_1,
    b_t_1,
    psi_1,
    dr_psi_1,
    dxi_psi_1,
    a2_2,
    nabla_a2_2,
    b_t_0_2,
    b_t_2,
    psi_2,
    dr_psi_2,
    dxi_psi_2,
    a2_3,
    nabla_a2_3,
    b_t_0_3,
    b_t_3,
    psi_3,
    dr_psi_3,
    dxi_psi_3,
    a2_4,
    nabla_a2_4,
    b_t_0_4,
    b_t_4,
    psi_4,
    dr_psi_4,
    dxi_psi_4,
):
    """
    Evolve the r and pr coordinates of plasma particles to the next xi step
    using a Runge-Kutta method of 4th order.

    Parameters
    ----------
    dxi, dr : float
        Longitudinal and radial step.
    xi : float
        Current longitudinal position of the plasma slice.
    r, pr, gamma, q : ndarray
        Radial position, radial momentum, Lorentz factor and charge of the
        plasma particles.
    r_max_plasma : float
        Maximum radial extent of the plasma
    dr_p : float
        Initial radial spacing between plasma particles.
    pc : float
        Coefficient for the parabolic radial plasma profile.
    a2, nabla_a2, b_t_0 : ndarray
        Laser and beam source fields.
    xi_fld, r_fld : ndarray
        Grid coordinates
    dr_1, ..., dr_4 : ndarray
        Arrays containing the derivative of the radial position of the
        particles at the current slice and the 3 intermediate steps.
    dpr_1, ..., dpr_4 : ndarray
        Arrays containing the derivative of the radial momentum of the
        particles at the current slice and the 3 intermediate steps.
    a2_i, ..., dxi_psi_i : ndarray
        Arrays where the field values at the particle positions at substep i
        will be stored.
    """
    # Calculate derivatives of r and pr at the current slice.
    calculate_derivatives(
        pr, gamma, b_t_0_1, nabla_a2_1, b_t_1, psi_1, dr_psi_1, dr_1, dpr_1
    )

    # Calculate derivatives of r and pr at the three RK4 substeps.
    derivatives_substep(
        xi - dxi * 0.5,
        r + dxi * dr_1 * 0.5,
        pr + dxi * dpr_1 * 0.5,
        q,
        dxi,
        dr,
        r_max_plasma,
        dr_p,
        pc,
        a2,
        nabla_a2,
        b_t_0,
        r_fld,
        xi_fld,
        a2_2,
        nabla_a2_2,
        b_t_0_2,
        b_t_2,
        psi_2,
        dr_psi_2,
        dxi_psi_2,
        dr_2,
        dpr_2,
    )
    derivatives_substep(
        xi - dxi * 0.5,
        r + dxi * dr_2 * 0.5,
        pr + dxi * dpr_2 * 0.5,
        q,
        dxi,
        dr,
        r_max_plasma,
        dr_p,
        pc,
        a2,
        nabla_a2,
        b_t_0,
        r_fld,
        xi_fld,
        a2_3,
        nabla_a2_3,
        b_t_0_3,
        b_t_3,
        psi_3,
        dr_psi_3,
        dxi_psi_3,
        dr_3,
        dpr_3,
    )
    derivatives_substep(
        xi - dxi,
        r + dxi * dr_3,
        pr + dxi * dpr_3,
        q,
        dxi,
        dr,
        r_max_plasma,
        dr_p,
        pc,
        a2,
        nabla_a2,
        b_t_0,
        r_fld,
        xi_fld,
        a2_4,
        nabla_a2_4,
        b_t_0_4,
        b_t_4,
        psi_4,
        dr_psi_4,
        dxi_psi_4,
        dr_4,
        dpr_4,
    )

    # Advance radial position and momentum.
    apply_rk4(r, dxi, dr_1, dr_2, dr_3, dr_4)
    apply_rk4(pr, dxi, dpr_1, dpr_2, dpr_3, dpr_4)

    # If a particle has crossed the axis, mirror it.
    idx_neg = np.where(r < 0.0)
    if idx_neg[0].size > 0:
        r[idx_neg] *= -1.0
        pr[idx_neg] *= -1.0


@njit_serial()
def derivatives_substep(
    xi,
    r,
    pr,
    q,
    dxi,
    dr,
    r_max_plasma,
    dr_p,
    pc,
    a2,
    nabla_a2,
    b_t_0,
    r_fld,
    xi_fld,
    a2_i,
    nabla_a2_i,
    b_t_0_i,
    b_t_i,
    psi_i,
    dr_psi_i,
    dxi_psi_i,
    dr_i,
    dpr_i,
):
    """Calculate r and pr derivatives at the i-th RK4 substep.

    The Runge-Kutta method of 4th order requires knowing the derivative (slope)
    of the variable to advance not only at the current location, but also at
    three other intermediate steps. This method takes care of calculating
    the derivatives of r and pr at the 3 required substeps. This includes
    gathering and computing all the fields at the location of each particle,
    which are required to calculate the derivatives.

    Parameters
    ----------
    xi : float
        Current longitudinal position of the plasma slice.
    r, pr, q : ndarray
        Radial position, radial momentum and charge of the plasma particles.
    dxi, dr : float
        Grid spacing.
    r_max_plasma : float
        Maximum radial extent of the plasma
    dr_p : float
        Initial radial spacing between plasma particles.
    pc : float
        Coefficient for the parabolic radial plasma profile.
    a2, nabla_a2, b_t_0 : ndarray
        Laser and beam source fields.
    xi_fld, r_fld : ndarray
        Grid coordinates
    a2_i, ..., dxi_psi_i : ndarray
        Arrays where the field values at the particle positions at substep i
        will be stored.
    dr_i, dpr_i : ndarray
        Arrays that will contain the derivative of the radial position and of
        the radial momentum of the particles at substep i.
    """

    # Check for particles with negative radial position and mirror them.
    idx_neg = np.where(r < 0.0)
    if idx_neg[0].size > 0:
        r[idx_neg] *= -1.0
        pr[idx_neg] *= -1.0

    # Gather source terms at position of plasma particles.
    gather_sources_qs_baxevanis(
        a2,
        nabla_a2,
        b_t_0,
        xi_fld[0],
        xi_fld[-1],
        r_fld[0],
        r_fld[-1],
        dxi,
        dr,
        r,
        xi,
        a2_i,
        nabla_a2_i,
        b_t_0_i,
    )

    # Get sorted particle indices
    idx = np.argsort(r)

    # Calculate wakefield potential and derivatives at plasma particles.
    calculate_psi_and_derivatives_at_particles(
        r, pr, q, idx, r_max_plasma, dr_p, pc, psi_i, dr_psi_i, dxi_psi_i
    )

    # Calculate gamma of plasma particles
    gamma = (1.0 + pr**2 + a2_i + (1.0 + psi_i) ** 2) / (2.0 * (1.0 + psi_i))

    # Calculate azimuthal magnetic field from the plasma at the location of
    # the plasma particles.
    calculate_b_theta_at_particles(
        r,
        pr,
        q,
        gamma,
        psi_i,
        dr_psi_i,
        dxi_psi_i,
        b_t_0_i,
        nabla_a2_i,
        idx,
        dr_p,
        b_t_i,
    )

    # Using the gathered/calculated fields, compute derivatives of r and pr
    # at the current slice.
    calculate_derivatives(
        pr, gamma, b_t_0_i, nabla_a2_i, b_t_i, psi_i, dr_psi_i, dr_i, dpr_i
    )

    # For particles which crossed the axis and were inverted, invert now
    # back the sign of the derivatives.
    if idx_neg[0].size > 0:
        dr_i[idx_neg] *= -1.0
        dpr_i[idx_neg] *= -1.0


@njit_serial()
def calculate_derivatives(pr, gamma, b_t_0, nabla_a2, b_t_bar, psi, dr_psi, dr, dpr):
    """
    Calculate the derivative of the radial position and the radial momentum
    of the plasma particles at the current slice.

    Parameters
    ----------
    pr, gamma : ndarray
        Arrays containing the radial momentum and Lorentz factor of the
        plasma particles.
    b_t_0 : ndarray
        Array containing the value of the azimuthal magnetic field from
        the beam distribution at the position of each plasma particle.
    nabla_a2 : ndarray
        Array containing the value of the gradient of the laser normalized
        vector potential at the position of each plasma particle.
    b_t_bar : ndarray
        Array containing the value of the azimuthal magnetic field from
        the plasma at the position of each plasma particle.
    psi, dr_psi : ndarray
        Arrays containing the wakefield potential and its radial derivative
        at the position of each plasma particle.
    dr, dpr : ndarray
        Arrays where the value of the derivatives of the radial position and
        radial momentum will be stored.
    """
    # Calculate derivatives of r and pr.
    for i in range(pr.shape[0]):
        psi_i = psi[i]
        dpr[i] = (
            gamma[i] * dr_psi[i] / (1.0 + psi_i)
            - b_t_bar[i]
            - b_t_0[i]
            - nabla_a2[i] / (2.0 * (1.0 + psi_i))
        )
        dr[i] = pr[i] / (1.0 + psi_i)


@njit_serial()
def apply_rk4(x, dt, kx_1, kx_2, kx_3, kx_4):
    """Apply the Runge-Kutta method of 4th order to evolve `x`

    Parameters
    ----------
    x : ndarray
        Array containing the variable to be advanced.
    dt : _type_
        Discretization step size.
    dx_1, dx_2, dx_3, dx_4 : ndarray
        Arrays containing the derivatives of `x` at the 4 substeps.
    """
    inv_6 = 1.0 / 6.0
    for i in range(x.shape[0]):
        x[i] += dt * (kx_1[i] + 2.0 * (kx_2[i] + kx_3[i]) + kx_4[i]) * inv_6



"""This module contains methods for data analysis"""

import os
from typing import Optional, List, Dict

import numpy as np
from h5py import File as H5File
import aptools.data_analysis.beam_diagnostics as bd

from wake_t.particles.particle_bunch import ParticleBunch


def analyze_bunch(
    bunch: ParticleBunch,
    n_slices: Optional[int] = 50,
    len_slice: Optional[float] = None,
) -> Dict:
    # perform analysis
    dist = bunch.prop_distance
    params_analysis = _get_distribution_parameters(
        bunch.x,
        bunch.y,
        bunch.xi,
        bunch.px,
        bunch.py,
        bunch.pz,
        bunch.q,
        n_slices,
        len_slice,
    )

    # store data
    bunch_params = _store_bunch_parameters_into_dict(dist, *params_analysis)
    return bunch_params


def analyze_bunch_list(
    bunch_list: List[ParticleBunch],
    n_slices: Optional[int] = 50,
    len_slice: Optional[float] = None,
) -> Dict:
    # preallocate arrays
    list_len = len(bunch_list)
    a_x = np.zeros(list_len)
    a_y = np.zeros(list_len)
    b_x = np.zeros(list_len)
    b_y = np.zeros(list_len)
    g_x = np.zeros(list_len)
    g_y = np.zeros(list_len)
    ene = np.zeros(list_len)
    dist = np.zeros(list_len)
    ene_sp = np.zeros(list_len)
    ene_sp_sl_avg = np.zeros(list_len)
    em_x = np.zeros(list_len)
    em_y = np.zeros(list_len)
    em_x_sl_avg = np.zeros(list_len)
    em_y_sl_avg = np.zeros(list_len)
    s_x = np.zeros(list_len)
    s_y = np.zeros(list_len)
    x_avg = np.zeros(list_len)
    y_avg = np.zeros(list_len)
    i_peak = np.zeros(list_len)
    theta_x = np.zeros(list_len)
    theta_y = np.zeros(list_len)
    s_z = np.zeros(list_len)
    q_tot = np.zeros(list_len)

    # perform analysis
    for i, bunch in enumerate(bunch_list):
        dist[i] = bunch.prop_distance
        params_analysis = _get_distribution_parameters(
            bunch.x,
            bunch.y,
            bunch.xi,
            bunch.px,
            bunch.py,
            bunch.pz,
            bunch.q,
            n_slices,
            len_slice,
        )
        (
            theta_x[i],
            theta_y[i],
            x_avg[i],
            y_avg[i],
            s_x[i],
            s_y[i],
            a_x[i],
            a_y[i],
            b_x[i],
            b_y[i],
            g_x[i],
            g_y[i],
            em_x[i],
            em_y[i],
            em_x_sl_avg[i],
            em_y_sl_avg[i],
            ene[i],
            ene_sp[i],
            ene_sp_sl_avg[i],
            i_peak[i],
            s_z[i],
            q_tot[i],
        ) = params_analysis

    # store into dictionary
    bunch_list_params = _store_bunch_parameters_into_dict(
        dist,
        theta_x,
        theta_y,
        x_avg,
        y_avg,
        s_x,
        s_y,
        a_x,
        a_y,
        b_x,
        b_y,
        g_x,
        g_y,
        em_x,
        em_y,
        em_x_sl_avg,
        em_y_sl_avg,
        ene,
        ene_sp,
        ene_sp_sl_avg,
        i_peak,
        s_z,
        q_tot,
    )
    return bunch_list_params


def save_parameters_to_file(bunch_params, folder_path, file_name):
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)
    file_path = os.path.join(folder_path, file_name + ".h5")
    with H5File(file_path, "w") as h5_file:
        for param_name, param_data in bunch_params.items():
            h5_file.create_dataset(param_name, data=param_data)


def read_parameters_from_file(file_path):
    bunch_params = {}
    with H5File(file_path, "r") as h5_file:
        for param in h5_file.keys():
            bunch_params[param] = h5_file.get(param).value
    return bunch_params


def save_bunch_to_file(bunch, folder_path, file_name):
    if not os.path.exists(folder_path):
        os.makedirs(folder_path)
    file_path = os.path.join(folder_path, file_name + ".h5")
    with H5File(file_path, "w") as h5_file:
        h5_file.create_dataset("x", data=bunch.x)
        h5_file.create_dataset("y", data=bunch.y)
        h5_file.create_dataset("xi", data=bunch.xi)
        h5_file.create_dataset("px", data=bunch.px)
        h5_file.create_dataset("py", data=bunch.py)
        h5_file.create_dataset("pz", data=bunch.pz)
        h5_file.create_dataset("q", data=bunch.q)
        h5_file.attrs["prop_dist"] = bunch.prop_distance


def _get_distribution_parameters(x, y, z, px, py, pz, q, n_slices, len_slice):
    a_x, b_x, g_x = bd.twiss_parameters(x, px, pz, py, w=q)
    a_y, b_y, g_y = bd.twiss_parameters(y, py, pz, px, w=q)
    ene = bd.mean_energy(px, py, pz, w=q)
    ene_sp = bd.relative_rms_energy_spread(px, py, pz, w=q)
    ene_sp_slice_params = bd.relative_rms_slice_energy_spread(
        z, px, py, pz, w=q, n_slices=n_slices, len_slice=len_slice
    )
    ene_sp_sl, sl_w, sl_e, ene_sp_sl_avg = ene_sp_slice_params
    em_x = bd.normalized_transverse_rms_emittance(x, px, py, pz, w=q)
    em_y = bd.normalized_transverse_rms_emittance(y, py, px, pz, w=q)
    emitt_x_slice_params = bd.normalized_transverse_rms_slice_emittance(
        z, x, px, py, pz, w=q, n_slices=n_slices, len_slice=len_slice
    )
    em_x_sl, sl_w, sl_e, em_x_sl_avg = emitt_x_slice_params
    emitt_y_slice_params = bd.normalized_transverse_rms_slice_emittance(
        z, y, py, px, pz, w=q, n_slices=n_slices, len_slice=len_slice
    )
    em_y_sl, sl_w, sl_e, em_y_sl_avg = emitt_y_slice_params
    s_z = bd.rms_length(z, w=q)
    s_x = bd.rms_size(x, w=q)
    s_y = bd.rms_size(y, w=q)
    x_avg = np.average(x, weights=q)
    y_avg = np.average(y, weights=q)
    px_avg = np.average(px, weights=q)
    py_avg = np.average(py, weights=q)
    theta_x = px_avg / ene
    theta_y = py_avg / ene
    i_peak = bd.peak_current(z, q, n_slices=n_slices, len_slice=len_slice)
    q_tot = np.sum(q)
    return (
        theta_x,
        theta_y,
        x_avg,
        y_avg,
        s_x,
        s_y,
        a_x,
        a_y,
        b_x,
        b_y,
        g_x,
        g_y,
        em_x,
        em_y,
        em_x_sl_avg,
        em_y_sl_avg,
        ene,
        ene_sp,
        ene_sp_sl_avg,
        i_peak,
        s_z,
        q_tot,
    )


def _store_bunch_parameters_into_dict(
    dist,
    theta_x,
    theta_y,
    x_avg,
    y_avg,
    s_x,
    s_y,
    a_x,
    a_y,
    b_x,
    b_y,
    g_x,
    g_y,
    em_x,
    em_y,
    em_x_sl_avg,
    em_y_sl_avg,
    ene,
    ene_sp,
    ene_sp_sl_avg,
    i_peak,
    s_z,
    q_tot,
):
    params_dict = {
        "prop_dist": dist,
        "theta_x": theta_x,
        "theta_y": theta_y,
        "x_avg": x_avg,
        "y_avg": y_avg,
        "sigma_x": s_x,
        "sigma_y": s_y,
        "sigma_z": s_z,
        "alpha_x": a_x,
        "alpha_y": a_y,
        "beta_x": b_x,
        "beta_y": b_y,
        "gamma_x": g_x,
        "gamma_y": g_y,
        "emitt_x": em_x,
        "emitt_y": em_y,
        "avg_slice_emitt_x": em_x_sl_avg,
        "avg_slice_emitt_y": em_y_sl_avg,
        "avg_ene": ene,
        "rel_ene_spread": ene_sp,
        "avg_slice_rel_ene_spread": ene_sp_sl_avg,
        "i_peak": i_peak,
        "q_tot": q_tot,
    }
    return params_dict



"""Contains the base class for all numerical fields."""

from typing import Optional, List

import numpy as np

from .base import Field
from wake_t.particles.particle_bunch import ParticleBunch


class NumericalField(Field):
    """Base class for all fields that are computed numerically.

    All numerical fields must have a time step `dt_update` which determines how
    often they will be updated (i.e., recalculated). Every time `update` is
    called, the internal time is advanced by one step and the field is
    recalculated. The class also provides methods for initializing and evolving
    field properties (e.g., a laser pulse envelope).

    Each subclass must implement the logic for calculating and gathering the
    field and, optionally, for initializing and evolving any field properties.

    Parameters
    ----------
    dt_update : float
        Update period (in seconds) of the field.
    openpmd_diag_supported : bool
        Whether openPMD diagnostics are supported by the field.
    force_even_updates : bool
        During tracking, it can happen that the total simulation time
        is not an integer multiple of `dt_update`, so that the last
        update is used for less time than the others. If set to True,
        this parameter will modify `dt_update` (making it smaller, never
        larger) so that the total tracking time is an integer multiple
        of `dt_update`. This makes sure also that the fields are
        updated one last time exactly at the end of the stage.
    """

    def __init__(
        self,
        dt_update: float,
        openpmd_diag_supported: Optional[bool] = False,
        force_even_updates: Optional[bool] = False,
    ) -> None:
        super().__init__(openpmd_diag_supported=openpmd_diag_supported)
        self.dt_update = dt_update
        self.force_even_updates = force_even_updates
        self.initialized = False

    def update(self, bunches: List[ParticleBunch]) -> None:
        """Update field to the next time step.

        Parameters
        ----------
        bunches : list
            List of ``ParticleBunch`` that can be used to recompute/update
            the fields.
        """
        if not self.initialized:
            self.initialize_properties(bunches)
        else:
            self.evolve_properties(bunches)
        self.calculate_field(bunches)

    def initialize_properties(self, bunches: List[ParticleBunch]) -> None:
        """Initialize field properties."""
        self.t = 0.0
        self._initialize_properties(bunches)
        self.initialized = True

    def evolve_properties(self, bunches: List[ParticleBunch]) -> None:
        """Evolve field properties."""
        self.t += self.dt_update
        self._evolve_properties(bunches)

    def calculate_field(self, bunches: List[ParticleBunch]) -> None:
        """Calculate field using the current properties and given bunches."""
        self._calculate_field(bunches)

    def adjust_dt(self, t_final: float) -> None:
        """Autoadjust the time step of the field update."""
        if self.force_even_updates:
            n_updates = np.ceil(t_final / self.dt_update)
            self.dt_update = t_final / n_updates

    def _initialize_properties(self, bunches):
        pass

    def _evolve_properties(self, bunches):
        pass

    def _calculate_field(self, bunches):
        raise NotImplementedError



import numpy as np
import scipy.constants as ct

from wake_t.fields.analytical_field import AnalyticalField
from wake_t.utilities.numba import prange


class FocusingBlowoutField(AnalyticalField):
    def __init__(self, density_function):
        self.density = density_function

        def e_x(x, y, xi, t, ex, k):
            for i in prange(x.shape[0]):
                ex[i] = ct.c * k[i] * x[i]

        def e_y(x, y, xi, t, ey, k):
            for i in prange(x.shape[0]):
                ey[i] = ct.c * k[i] * y[i]

        super().__init__(e_x=e_x, e_y=e_y)

    def _pre_gather(self, x, y, xi, t):
        z = t * ct.c + xi
        n_p = self.density(z, 0.0)
        w_p = np.sqrt(n_p * ct.e**2 / (ct.m_e * ct.epsilon_0))
        k = (ct.m_e / (2 * ct.e * ct.c)) * w_p**2
        self.constants = k



"""This module contains custom definitions of the numba decorators."""

import os

from numba import (
    njit,
    __version__ as numba_version,
    set_num_threads,
    get_num_threads,
    prange,
)

if numba_version == "0.57.0":
    raise RuntimeError(
        "Wake-T is incompatible with numba 0.57.0.\n"
        "Please install either a later or an earlier version."
    )


# Check if the environment variable WAKET_DISABLE_CACHING is set to 1
# and in that case, disable caching
caching = True
if "WAKET_DISABLE_CACHING" in os.environ:
    if int(os.environ["WAKET_DISABLE_CACHING"]) == 1:
        caching = False


# Check if the environment variable WAKET_DISABLE_CACHING is set to 1
# and in that case, disable caching
num_threads = 1
if "WAKET_NUM_THREADS" in os.environ:
    num_threads = int(os.environ["WAKET_NUM_THREADS"])


# Define custom njit decorator for serial methods.
def njit_serial(*args, **kwargs):
    return njit(*args, cache=caching, **kwargs)


# Define custom njit decorator for parallel methods.
def njit_parallel(*args, **kwargs):
    return njit(*args, cache=caching, parallel=True, **kwargs)


__all__ = [
    "njit_serial",
    "njit_parallel",
    "num_threads",
    "set_num_threads",
    "get_num_threads",
    "prange",
]



"""Contains the definition of the `PlasmaParticles` class."""

from typing import Optional, List, Callable

import numpy as np
import scipy.constants as ct

from .psi_and_derivatives import (
    calculate_psi_with_interpolation,
    calculate_psi_and_derivatives_at_particles,
)
from .deposition import deposit_plasma_particles
from .gather import gather_bunch_sources, gather_laser_sources
from .b_theta import (
    calculate_b_theta_at_particles,
    calculate_b_theta_with_interpolation,
)
from .plasma_push.ab2 import evolve_plasma_ab2
from .utils import (
    calculate_chi,
    calculate_rho,
    update_gamma_and_pz,
    sort_particle_arrays,
    check_gamma,
    log,
)


class PlasmaParticles:
    """
    Class containing a 1D slice of plasma particles.

    In the current implementation, this class stores both the plasma electrons
    and ions. It would be useful to change this in the future so that it
    stores only a single species. This would allow us to more easily
    extend the wakefield model to cases with more than 2 species, which would
    be great to model ionization, for example.

    Parameters
    ----------
    r_max : float
        Maximum radial extension of the simulation box in normalized units.
    r_max_plasma : float
        Maximum radial extension of the plasma column in normalized units.
    dr : float
        Radial step size of the discretized simulation box.
    ppc : float
        Number of particles per cell.
    nr, nz : int
        Number of grid elements along `r` and `z`.
    radial_density : callable
        Function defining the radial density profile.
    max_gamma : float, optional
        Plasma particles whose ``gamma`` exceeds ``max_gamma`` are
        considered to violate the quasistatic condition and are put at
        rest (i.e., ``gamma=1.``, ``pr=pz=0.``). By default 10.
    ion_motion : bool, optional
        Whether to allow the plasma ions to move. By default, False.
    ion_mass : float, optional
        Mass of the plasma ions. By default, the mass of a proton.
    free_electrons_per_ion : int, optional
        Number of free electrons per ion. The ion charge is adjusted
        accordingly to maintain a quasi-neutral plasma (i.e.,
        ion charge = e * free_electrons_per_ion). By default, 1.
    pusher : str, optional
        The pusher used to evolve the plasma particles. Possible values
        are ``'ab2'`` (Adams-Bashforth 2nd order).
    shape : str
        Particle shape to be used for the beam charge deposition. Possible
        values are 'linear' or 'cubic'. By default 'linear'.
    store_history : bool, optional
        Whether to store the plasma particle evolution. This might be needed
        for diagnostics or because of the use of adaptive grids. By default,
        ``False``.
    diags : list, optional
        List of particle quantities to save to diagnostics.
    """

    def __init__(
        self,
        r_max: float,
        r_max_plasma: float,
        dr: float,
        ppc: float,
        nr: int,
        nz: int,
        radial_density: Callable[[float], float],
        max_gamma: Optional[float] = 10.0,
        ion_motion: Optional[bool] = True,
        ion_mass: Optional[float] = ct.m_p,
        free_electrons_per_ion: Optional[int] = 1,
        pusher: Optional[str] = "ab2",
        shape: Optional[str] = "linear",
        store_history: Optional[bool] = False,
        diags: Optional[List[str]] = [],
    ):
        # Store parameters.
        self.r_max = r_max
        self.r_max_plasma = r_max_plasma
        self.radial_density = radial_density
        self.dr = dr
        self.ppc = ppc
        self.pusher = pusher
        self.shape = shape
        self.max_gamma = max_gamma
        self.nr = nr
        self.nz = nz
        self.ion_motion = ion_motion
        self.ion_mass = ion_mass
        self.free_electrons_per_ion = free_electrons_per_ion
        self.store_history = store_history
        self.diags = diags

    def initialize(self):
        """Initialize column of plasma particles."""

        # Create radial distribution of plasma particles.
        rmin = 0.0
        for i in range(self.ppc.shape[0]):
            rmax = self.ppc[i, 0]
            ppc = self.ppc[i, 1]

            n_elec = int(np.round((rmax - rmin) / self.dr * ppc))
            dr_p_i = self.dr / ppc
            rmax = rmin + n_elec * dr_p_i

            r_i = np.linspace(rmin + dr_p_i / 2, rmax - dr_p_i / 2, n_elec)
            dr_p_i = np.ones(n_elec) * dr_p_i
            if i == 0:
                r = r_i
                dr_p = dr_p_i
            else:
                r = np.concatenate((r, r_i))
                dr_p = np.concatenate((dr_p, dr_p_i))

            rmin = rmax

        # Determine number of particles.
        self.n_elec = r.shape[0]
        self.n_part = self.n_elec * 2

        # Initialize particle arrays.
        # `q_center` represents the charge until the particle center. That is,
        # the charge of the first half of the particle.
        pr = np.zeros(self.n_elec)
        pz = np.zeros(self.n_elec)
        gamma = np.ones(self.n_elec)
        id = np.arange(self.n_elec, dtype=np.int32)
        w = dr_p * r * self.radial_density(r)
        w_center = w / 2 - dr_p**2 / 8

        # Charge and mass of the macroparticles of each species.
        self.m_elec = self.free_electrons_per_ion
        self.m_ion = self.ion_mass / ct.m_e
        self.q_species_elec = self.free_electrons_per_ion
        self.q_species_ion = -self.free_electrons_per_ion

        # Combine arrays of both species.
        self.r = np.concatenate((r, r))
        self.dr_p = np.concatenate((dr_p, dr_p))
        self.pr = np.concatenate((pr, pr))
        self.pz = np.concatenate((pz, pz))
        self.gamma = np.concatenate((gamma, gamma))
        self.w = np.concatenate((w, w))
        self.w_center = np.concatenate((w_center, w_center))
        self.r_to_x = np.ones(self.n_part, dtype=np.int32)
        self.id = np.concatenate((id, id))

        # Create history arrays.
        if self.store_history:
            self.r_hist = np.zeros((self.nz, self.n_part))
            self.log_r_hist = np.zeros((self.nz, self.n_part))
            self.xi_hist = np.zeros((self.nz, self.n_part))
            self.pr_hist = np.zeros((self.nz, self.n_part))
            self.pz_hist = np.zeros((self.nz, self.n_part))
            self.w_hist = np.zeros((self.nz, self.n_part))
            self.r_to_x_hist = np.zeros((self.nz, self.n_part), dtype=np.int32)
            self.id_hist = np.zeros((self.nz, self.n_part), dtype=np.int32)
            self.sum_1_hist = np.zeros((self.nz, self.n_part + 2))
            self.sum_2_hist = np.zeros((self.nz, self.n_part + 2))
            self.a_i_hist = np.zeros((self.nz, self.n_elec))
            self.b_i_hist = np.zeros((self.nz, self.n_elec))
            self.a_0_hist = np.zeros(self.nz)
            self.i_push = 0
            self.xi_current = 0.0

        self.ions_computed = False

        # Allocate arrays that will contain the fields experienced by the
        # particles.
        self._allocate_field_arrays()
        self._make_species_views()

        # Allocate arrays needed for the particle pusher.
        if self.pusher == "ab2":
            self._allocate_ab2_arrays()

    def sort(self):
        """Sort plasma particles radially.

        The `q_species` and `m` arrays do not need to be sorted because all
        particles have the same value.
        """
        i_sort_e = np.argsort(self.r_elec, kind="stable")
        sort_particle_arrays(
            self.r_elec,
            self.dr_p_elec,
            self.pr_elec,
            self.pz_elec,
            self.gamma_elec,
            self.w_elec,
            self.w_center_elec,
            self.r_to_x_elec,
            self.id_elec,
            self._dr_e,
            self._dpr_e,
            i_sort_e,
        )
        if self.ion_motion:
            i_sort_i = np.argsort(self.r_ion, kind="stable")
            sort_particle_arrays(
                self.r_ion,
                self.dr_p_ion,
                self.pr_ion,
                self.pz_ion,
                self.gamma_ion,
                self.w_ion,
                self.w_center_ion,
                self.r_to_x_ion,
                self.id_ion,
                self._dr_i,
                self._dpr_i,
                i_sort_i,
            )

    def gather_laser_sources(self, a2, nabla_a2, r_min, r_max, dr):
        """Gather the source terms (a^2 and nabla(a)^2) from the laser."""
        if self.ion_motion:
            gather_laser_sources(
                a2,
                nabla_a2,
                r_min,
                r_max,
                dr,
                self.r,
                self._a2,
                self._nabla_a2,
            )
        else:
            gather_laser_sources(
                a2,
                nabla_a2,
                r_min,
                r_max,
                dr,
                self.r_elec,
                self._a2_e,
                self._nabla_a2_e,
            )

    def gather_bunch_sources(
        self, source_arrays, source_xi_indices, source_metadata, slice_i
    ):
        """Gather the source terms (b_theta) from the particle bunches."""
        self._b_t_0[:] = 0.0
        for i in range(len(source_arrays)):
            array = source_arrays[i]
            idx = source_xi_indices[i]
            md = source_metadata[i]
            r_min = md[0]
            r_max = md[1]
            dr = md[2]
            if slice_i in idx:
                xi_index = slice_i + 2 - idx[0]
                if self.ion_motion:
                    gather_bunch_sources(
                        array[xi_index], r_min, r_max, dr, self.r, self._b_t_0
                    )
                else:
                    gather_bunch_sources(
                        array[xi_index],
                        r_min,
                        r_max,
                        dr,
                        self.r_elec,
                        self._b_t_0_e,
                    )

    def calculate_fields(self):
        """Calculate the fields at the plasma particles."""
        # Precalculate logarithms (expensive) to avoid doing so several times.
        log(self.r_elec, self.log_r_elec)
        if self.ion_motion or not self.ions_computed:
            log(self.r_ion, self.log_r_ion)

        calculate_psi_and_derivatives_at_particles(
            self.r_elec,
            self.log_r_elec,
            self.pr_elec,
            self.w_elec,
            self.w_center_elec,
            self.q_species_elec,
            self.r_ion,
            self.log_r_ion,
            self.pr_ion,
            self.w_ion,
            self.w_center_ion,
            self.q_species_ion,
            self.ion_motion,
            self.ions_computed,
            self._sum_1_e,
            self._sum_2_e,
            self._sum_3_e,
            self._sum_1_i,
            self._sum_2_i,
            self._sum_3_i,
            self._psi_e,
            self._dr_psi_e,
            self._dxi_psi_e,
            self._psi_i,
            self._dr_psi_i,
            self._dxi_psi_i,
            self._psi,
            self._dr_psi,
            self._dxi_psi,
        )
        update_gamma_and_pz(
            self.gamma_elec,
            self.pz_elec,
            self.pr_elec,
            self._a2_e,
            self._psi_e,
            self.q_species_elec,
            self.m_elec,
        )
        if self.ion_motion:
            update_gamma_and_pz(
                self.gamma_ion,
                self.pz_ion,
                self.pr_ion,
                self._a2_i,
                self._psi_i,
                self.q_species_ion,
                self.m_ion,
            )
        check_gamma(self.gamma_elec, self.pz_elec, self.pr_elec, self.max_gamma)
        calculate_b_theta_at_particles(
            self.r_elec,
            self.pr_elec,
            self.w_elec,
            self.w_center_elec,
            self.gamma_elec,
            self.q_species_elec,
            self.r_ion,
            self.ion_motion,
            self._psi_e,
            self._dr_psi_e,
            self._dxi_psi_e,
            self._b_t_0_e,
            self._nabla_a2_e,
            self._A,
            self._B,
            self._C,
            self._K,
            self._U,
            self._a_0,
            self._a_i,
            self._b_i,
            self._b_t_e,
            self._b_t_i,
        )

    def calculate_psi_at_grid(self, r_eval, psi):
        """Calculate psi on the current grid slice."""
        calculate_psi_with_interpolation(
            r_eval,
            self.r_elec,
            self.log_r_elec,
            self._sum_1_e,
            self._sum_2_e,
            psi,
        )
        calculate_psi_with_interpolation(
            r_eval,
            self.r_ion,
            self.log_r_ion,
            self._sum_1_i,
            self._sum_2_i,
            psi,
            add=True,
        )

    def calculate_b_theta_at_grid(self, r_eval, b_theta):
        """Calculate b_theta on the current grid slice."""
        calculate_b_theta_with_interpolation(
            r_eval, self._a_0[0], self._a_i, self._b_i, self.r_elec, b_theta
        )

    def evolve(self, dxi):
        """Evolve plasma particles to next longitudinal slice."""
        evolve_plasma_ab2(
            dxi,
            self.r_elec,
            self.pr_elec,
            self.gamma_elec,
            self.m_elec,
            self.q_species_elec,
            self.r_to_x_elec,
            self._nabla_a2_e,
            self._b_t_0_e,
            self._b_t_e,
            self._psi_e,
            self._dr_psi_e,
            self._dr_e,
            self._dpr_e,
        )
        if self.ion_motion:
            evolve_plasma_ab2(
                dxi,
                self.r_ion,
                self.pr_ion,
                self.gamma_ion,
                self.m_ion,
                self.q_species_ion,
                self.r_to_x_ion,
                self._nabla_a2_i,
                self._b_t_0_i,
                self._b_t_i,
                self._psi_i,
                self._dr_psi_i,
                self._dr_i,
                self._dpr_i,
            )

        if self.store_history:
            self.i_push += 1
            self.xi_current -= dxi
            self._move_auxiliary_arrays_to_next_slice()

    def calculate_weights(self):
        """Calculate the plasma density weights of each particle."""
        calculate_rho(
            self.q_species_elec,
            self.w_elec,
            self.pz_elec,
            self.gamma_elec,
            self._rho_e,
        )
        if self.ion_motion or not self.ions_computed:
            calculate_rho(
                self.q_species_ion,
                self.w_ion,
                self.pz_ion,
                self.gamma_ion,
                self._rho_i,
            )

    def deposit_rho(self, rho, rho_e, rho_i, r_fld, nr, dr):
        """Deposit plasma density on a grid slice."""
        self.calculate_weights()
        # Deposit electrons
        deposit_plasma_particles(
            self.r_elec, self._rho_e, r_fld[0], nr, dr, rho_e, self.shape
        )

        # Deposit ions
        deposit_plasma_particles(
            self.r_ion, self._rho_i, r_fld[0], nr, dr, rho_i, self.shape
        )
        rho[:] = rho_e
        rho += rho_i

    def deposit_chi(self, chi, r_fld, nr, dr):
        """Deposit plasma susceptibility on a grid slice."""
        calculate_chi(
            self.q_species_elec,
            self.w_elec,
            self.pz_elec,
            self.gamma_elec,
            self._chi_e,
        )
        deposit_plasma_particles(
            self.r_elec, self._chi_e, r_fld[0], nr, dr, chi, self.shape
        )

    def get_history(self):
        """Get the history of the evolution of the plasma particles.

        Returns
        -------
        dict
            A dictionary containing the particle history arrays.
        """
        if self.store_history:
            history = {
                "r_hist": self.r_hist,
                "log_r_hist": self.log_r_hist,
                "xi_hist": self.xi_hist,
                "pr_hist": self.pr_hist,
                "pz_hist": self.pz_hist,
                "w_hist": self.w_hist,
                "r_to_x_hist": self.r_to_x_hist,
                "id_hist": self.id_hist,
                "sum_1_hist": self.sum_1_hist,
                "sum_2_hist": self.sum_2_hist,
                "a_i_hist": self.a_i_hist,
                "b_i_hist": self.b_i_hist,
                "a_0_hist": self.a_0_hist,
            }
            return history

    def store_current_step(self):
        """Store current particle properties in the history arrays."""
        if "r" in self.diags or self.store_history:
            self.r_hist[-1 - self.i_push] = self.r
        if "z" in self.diags:
            self.xi_hist[-1 - self.i_push] = self.xi_current
        if "pr" in self.diags:
            self.pr_hist[-1 - self.i_push] = self.pr
        if "pz" in self.diags:
            self.pz_hist[-1 - self.i_push] = self.pz
        if "w" in self.diags:
            self.w_hist[-1 - self.i_push] = self._rho
        if "r_to_x" in self.diags:
            self.r_to_x_hist[-1 - self.i_push] = self.r_to_x
        if "id" in self.diags:
            self.id_hist[-1 - self.i_push] = self.id
        if self.store_history:
            self.a_0_hist[-1 - self.i_push] = self._a_0[0]

    def _allocate_field_arrays(self):
        """Allocate arrays for the fields experienced by the particles.

        In order to evolve the particles to the next longitudinal position,
        it is necessary to know the fields that they are experiencing. These
        arrays are used for storing the value of these fields at the location
        of each particle.
        """
        # When storing the particle history, define the following auxiliary
        # arrays as views of a 1D slice of the history arrays.
        if self.store_history:
            self._a_i = self.a_i_hist[-1]
            self._b_i = self.b_i_hist[-1]
            self._sum_1 = self.sum_1_hist[-1]
            self._sum_2 = self.sum_2_hist[-1]
            self._rho = self.w_hist[-1]
            self._log_r = self.log_r_hist[-1]
        else:
            self._a_i = np.zeros(self.n_elec)
            self._b_i = np.zeros(self.n_elec)
            self._sum_1 = np.zeros(self.n_part + 2)
            self._sum_2 = np.zeros(self.n_part + 2)
            self._rho = np.zeros(self.n_part)
            self._log_r = np.zeros(self.n_part)

        self._a2 = np.zeros(self.n_part)
        self._nabla_a2 = np.zeros(self.n_part)
        self._b_t_0 = np.zeros(self.n_part)
        self._b_t = np.zeros(self.n_part)
        self._psi = np.zeros(self.n_part)
        self._dr_psi = np.zeros(self.n_part)
        self._dxi_psi = np.zeros(self.n_part)
        self._chi = np.zeros(self.n_part)
        self._sum_3_e = np.zeros(self.n_elec + 1)
        self._sum_3_i = np.zeros(self.n_elec + 1)
        self._a_0 = np.zeros(1)
        self._A = np.zeros(self.n_elec)
        self._B = np.zeros(self.n_elec)
        self._C = np.zeros(self.n_elec)
        self._K = np.zeros(self.n_elec)
        self._U = np.zeros(self.n_elec)

    def _make_species_views(self):
        """Make species arrays as partial views of the particle arrays."""
        self.r_elec = self.r[: self.n_elec]
        self.log_r_elec = self._log_r[: self.n_elec]
        self.dr_p_elec = self.dr_p[: self.n_elec]
        self.pr_elec = self.pr[: self.n_elec]
        self.pz_elec = self.pz[: self.n_elec]
        self.gamma_elec = self.gamma[: self.n_elec]
        self.w_elec = self.w[: self.n_elec]
        self.w_center_elec = self.w_center[: self.n_elec]
        self.r_to_x_elec = self.r_to_x[: self.n_elec]
        self.id_elec = self.id[: self.n_elec]

        self.r_ion = self.r[self.n_elec :]
        self.log_r_ion = self._log_r[self.n_elec :]
        self.dr_p_ion = self.dr_p[self.n_elec :]
        self.pr_ion = self.pr[self.n_elec :]
        self.pz_ion = self.pz[self.n_elec :]
        self.gamma_ion = self.gamma[self.n_elec :]
        self.w_ion = self.w[self.n_elec :]
        self.w_center_ion = self.w_center[self.n_elec :]
        self.r_to_x_ion = self.r_to_x[self.n_elec :]
        self.id_ion = self.id[self.n_elec :]

        self._psi_e = self._psi[: self.n_elec]
        self._dr_psi_e = self._dr_psi[: self.n_elec]
        self._dxi_psi_e = self._dxi_psi[: self.n_elec]
        self._psi_i = self._psi[self.n_elec :]
        self._dr_psi_i = self._dr_psi[self.n_elec :]
        self._dxi_psi_i = self._dxi_psi[self.n_elec :]
        self._b_t_e = self._b_t[: self.n_elec]
        self._b_t_i = self._b_t[self.n_elec :]
        self._b_t_0_e = self._b_t_0[: self.n_elec]
        self._b_t_0_i = self._b_t_0[self.n_elec :]
        self._nabla_a2_e = self._nabla_a2[: self.n_elec]
        self._nabla_a2_i = self._nabla_a2[self.n_elec :]
        self._a2_e = self._a2[: self.n_elec]
        self._a2_i = self._a2[self.n_elec :]
        self._sum_1_e = self._sum_1[: self.n_elec + 1]
        self._sum_2_e = self._sum_2[: self.n_elec + 1]
        self._sum_1_i = self._sum_1[self.n_elec + 1 :]
        self._sum_2_i = self._sum_2[self.n_elec + 1 :]
        self._rho_e = self._rho[: self.n_elec]
        self._rho_i = self._rho[self.n_elec :]
        self._chi_e = self._chi[: self.n_elec]
        self._chi_i = self._chi[self.n_elec :]

    def _allocate_ab2_arrays(self):
        """Allocate the arrays needed for the 5th order Adams-Bashforth pusher.

        The AB2 pusher needs the derivatives of r and pr for each particle
        at the last 2 plasma slices. This method allocates the arrays that will
        store these derivatives.
        """
        if self.ion_motion:
            size = self.n_part
        else:
            size = self.n_elec
        self._dr = np.zeros((2, size))
        self._dpr = np.zeros((2, size))
        self._dr_e = self._dr[:, : self.n_elec]
        self._dpr_e = self._dpr[:, : self.n_elec]
        self._dr_i = self._dr[:, self.n_elec :]
        self._dpr_i = self._dpr[:, self.n_elec :]

    def _move_auxiliary_arrays_to_next_slice(self):
        """Point auxiliary 1D arrays to next slice of the 2D history arrays.

        When storing the particle history, some auxiliary arrays (e.g., those
        storing the cumulative sums, the a_i, b_i coefficients, ...) have to be
        stored at every longitudinal step. In principle, this used to be done
        by writing the 1D auxiliary arrays into the corresponding slice of the
        2D history arrays. However, this is time consuming as it leads to
        copying data at every step. In order to avoid this, the auxiliary
        arrays are defined simply as views of a 1D slice of the history arrays
        so that the data is written directly to the history without it being a
        copy. In order to make this work, the slice to which the auxiliary
        arrays point to needs to be moved at each step. This is what this
        method does.
        """
        self._a_i = self.a_i_hist[-1 - self.i_push]
        self._b_i = self.b_i_hist[-1 - self.i_push]
        self._sum_1 = self.sum_1_hist[-1 - self.i_push]
        self._sum_2 = self.sum_2_hist[-1 - self.i_push]
        self._rho = self.w_hist[-1 - self.i_push]
        self._log_r = self.log_r_hist[-1 - self.i_push]

        self._sum_1_e = self._sum_1[: self.n_elec + 1]
        self._sum_2_e = self._sum_2[: self.n_elec + 1]
        self._sum_1_i = self._sum_1[self.n_elec + 1 :]
        self._sum_2_i = self._sum_2[self.n_elec + 1 :]
        self._rho_e = self._rho[: self.n_elec]
        self._rho_i = self._rho[self.n_elec :]
        self.log_r_elec = self._log_r[: self.n_elec]
        self.log_r_ion = self._log_r[self.n_elec :]

        if not self.ion_motion:
            self._sum_1_i[:] = self.sum_1_hist[-self.i_push, self.n_elec + 1 :]
            self._sum_2_i[:] = self.sum_2_hist[-self.i_push, self.n_elec + 1 :]
            self.log_r_ion[:] = self.log_r_hist[-self.i_push, self.n_elec :]



"""Contains the definition of the `AdaptiveGrid` class."""

from typing import Optional

import numpy as np
import scipy.constants as ct
import aptools.plasma_accel.general_equations as ge

from wake_t.utilities.numba import njit_serial
from wake_t.particles.interpolation import gather_main_fields_cyl_linear
from .psi_and_derivatives import calculate_psi_with_interpolation
from .b_theta import calculate_b_theta_with_interpolation
from .b_theta_bunch import calculate_bunch_source, deposit_bunch_charge
from .utils import longitudinal_gradient, radial_gradient


class AdaptiveGrid:
    """Grid whose size dynamically adapts to the extent of a particle bunch.

    The number of radial cells is fixed, but its transverse extent is
    continually adapted to fit the whole particle distribution (unless
    a fixed value for `r_max` is given).

    The longitudinal grid spacing is always constant, and set to the
    longitudinal step of the plasma particles. The longitudinal
    extent and number of grid points are always adjusted so that the whole
    particle bunch fits within the grid.

    Parameters
    ----------
    x, y, xi : ndarray
        The transverse and longitudinal coordinates of the bunch particles.
    bunch_name : str
        The name of the bunch that is being covered by the grid.
    nr, nxi : int
        Radial and longitudinal resolution of the grid.
    xi_plasma : ndarray
        Array containing the possible longitudinal locations of the plasma
        particles.
    r_max : float, optional
        Radial extent of the grid. If not given, the radial extent is
        dynamically updated with the beam size. If given, the radial extent
        is always fixed to the specified value.
    r_lim : float, optional
        Limit to the radial extent of the grid. If not given, the radial extent
        is dynamically updated with the beam size to fit the whole beam.
        If given, the radial extent will never be larger than the specified
        value. Bunch particles that escape the grid transversely with
        deposit to and gather from the base grid (if they haven't escaped
        from it too).
    """

    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        xi: np.ndarray,
        bunch_name: str,
        nr: int,
        nxi: int,
        xi_plasma: np.ndarray,
        r_max: Optional[float] = None,
        r_lim: Optional[float] = None,
    ):
        self.bunch_name = bunch_name
        self.xi_plasma = xi_plasma
        self.dxi = (xi_plasma[-1] - xi_plasma[0]) / (nxi - 1)
        self.nr_border = 2
        self.nxi_border = 2
        self.nr = nr + self.nr_border
        self._r_max = r_max
        self._r_lim = r_lim
        self._r_max_hist = []

        self._update(x, y, xi)

    @property
    def r_min_cell(self):
        """Radial position of first cell (ignoring guard cells)."""
        return self.r_grid[0]

    @property
    def r_max_cell(self):
        """Radial position of last cell (ignoring guard and border cells)."""
        return self.r_grid[-1 - self.nr_border]

    @property
    def r_max(self):
        """Radial extent of the grid, ignoring guard and border cells."""
        return self.r_max_cell + 0.5 * self.dr

    @property
    def r_max_cell_guard(self):
        """Radial position of last guard grid cell."""
        return self.r_grid[-1] + 2 * self.dr

    def update_if_needed(self, x, y, xi, n_p, pp_hist):
        """
        Update grid size and fields if bunch extent has changed sufficiently.

        Parameters
        ----------
        x, y, xi : ndarray
            The transverse and longitudinal coordinates of the bunch particles.
        n_p : float
            The plasma density.
        pp_hist : dict
            Dictionary containing arrays with the history of the plasma
            particles.
        """
        update_r = False
        # Only trigger radial update if the radial size is not fixed.
        if self._r_max is None:
            r_max_beam = np.max(np.sqrt(x**2 + y**2))
            update_r = (r_max_beam > self.r_max_cell) or (
                r_max_beam < self.r_max_cell * 0.9
            )
            # It a radial limit is set, update only if limit has not been
            # reached.
            if update_r and self._r_lim is not None:
                if r_max_beam < self._r_lim:
                    update_r = True
                elif self.r_max_cell != self._r_lim:
                    update_r = True
                else:
                    update_r = False

        xi_min_beam = np.min(xi)
        xi_max_beam = np.max(xi)
        update_xi = (xi_min_beam < self.xi_grid[0 + self.nxi_border]) or (
            xi_max_beam > self.xi_grid[-1 - self.nxi_border]
        )
        if update_r or update_xi:
            self._update(x, y, xi)
            self.calculate_fields(n_p, pp_hist, reset_fields=False)

    def calculate_fields(self, n_p, pp_hist, reset_fields=True):
        """Calculate the E and B fields from the plasma at the grid.

        Parameters
        ----------
        n_p : float
            The plasma density.
        pp_hist : dict
            Dictionary containing arrays with the history of the plasma
            particles.
        reset_fields : bool
            Whether the fields should be reset to zero before calculating.
        """
        if reset_fields:
            self._reset_fields()
        s_d = ge.plasma_skin_depth(n_p * 1e-6)
        calculate_fields_on_grid(
            self.i_grid,
            self.r_grid,
            s_d,
            self.psi_grid,
            self.b_t,
            pp_hist["r_hist"],
            pp_hist["log_r_hist"],
            pp_hist["sum_1_hist"],
            pp_hist["sum_2_hist"],
            pp_hist["a_0_hist"],
            pp_hist["a_i_hist"],
            pp_hist["b_i_hist"],
        )

        E_0 = ge.plasma_cold_non_relativisct_wave_breaking_field(n_p * 1e-6)
        longitudinal_gradient(
            self.psi_grid[2:-2, 2:-2], self.dxi / s_d, self.e_z[2:-2, 2:-2]
        )
        radial_gradient(self.psi_grid[2:-2, 2:-2], self.dr / s_d, self.e_r[2:-2, 2:-2])
        self.e_r -= self.b_t
        self.e_z *= -E_0
        self.e_r *= -E_0
        self.b_t *= E_0 / ct.c

    def calculate_bunch_source(self, bunch, n_p, p_shape):
        """Calculate the source term (B_theta) of the bunch within the grid.

        Parameters
        ----------
        bunch : ParticleBunch
            The particle bunch.
        n_p : float
            The plasma density.
        p_shape : str
            The particle shape.
        """
        self.b_t_bunch[:] = 0.0
        self.q_bunch[:] = 0.0
        all_deposited = deposit_bunch_charge(
            bunch.x,
            bunch.y,
            bunch.xi,
            bunch.q,
            n_p,
            self.nr - self.nr_border,
            self.nxi,
            self.r_grid,
            self.xi_grid,
            self.dr,
            self.dxi,
            p_shape,
            self.q_bunch,
        )
        calculate_bunch_source(self.q_bunch, self.nr, self.nxi, self.b_t_bunch)
        return all_deposited

    def gather_fields(self, x, y, z, ex, ey, ez, bx, by, bz):
        """Gather the plasma fields at the location of the bunch particles.

        Parameters
        ----------
        x, y, z : ndarray
            The transverse and longitudinal coordinates of the bunch particles.
        ex, ey, ez, bx, by, bz : ndarray
            The arrays where the gathered field components will be stored.
        """
        return gather_main_fields_cyl_linear(
            self.e_r,
            self.e_z,
            self.b_t,
            self.xi_min,
            self.xi_max,
            self.r_min_cell,
            self.r_max_cell,
            self.dxi,
            self.dr,
            x,
            y,
            z,
            ex,
            ey,
            ez,
            bx,
            by,
            bz,
        )

    def get_openpmd_data(self, global_time, diags):
        """Get the field data at the grid to store in the openpmd diagnostics.

        Parameters
        ----------
        global_time : float
            The global time of the simulation.
        diags : list
            A list of strings with the names of the fields to include in the
            diagnostics.

        Returns
        -------
        dict
        """

        # Grid parameters.
        grid_spacing = [self.dr, self.dxi]
        grid_labels = ["r", "z"]
        grid_global_offset = [0.0, global_time * ct.c + self.xi_min]

        # Initialize field diags lists.
        names = []
        comps = []
        attrs = []
        arrays = []

        # Add requested fields to lists.
        if "E" in diags:
            names += ["E"]
            comps += [["r", "z"]]
            attrs += [{}]
            arrays += [
                [
                    np.ascontiguousarray(self.e_r.T[2:-2, 2:-2]),
                    np.ascontiguousarray(self.e_z.T[2:-2, 2:-2]),
                ]
            ]
        if "B" in diags:
            names += ["B"]
            comps += [["t"]]
            attrs += [{}]
            arrays += [[np.ascontiguousarray(self.b_t.T[2:-2, 2:-2])]]

        # Create dictionary with all diagnostics data.
        comp_pos = [[0.5, 0.0]] * len(names)
        fld_zip = zip(names, comps, attrs, arrays, comp_pos)
        diag_data = {}
        diag_data["fields"] = []
        for fld, comps, attrs, arrays, pos in fld_zip:
            fld += "_" + self.bunch_name
            diag_data["fields"].append(fld)
            diag_data[fld] = {}
            if comps is not None:
                diag_data[fld]["comps"] = {}
                for comp, arr in zip(comps, arrays):
                    diag_data[fld]["comps"][comp] = {}
                    diag_data[fld]["comps"][comp]["array"] = arr
                    diag_data[fld]["comps"][comp]["position"] = pos
            else:
                diag_data[fld]["array"] = arrays[0]
                diag_data[fld]["position"] = pos
            diag_data[fld]["grid"] = {}
            diag_data[fld]["grid"]["spacing"] = grid_spacing
            diag_data[fld]["grid"]["labels"] = grid_labels
            diag_data[fld]["grid"]["global_offset"] = grid_global_offset
            diag_data[fld]["attributes"] = attrs

        return diag_data

    def _update(self, x, y, xi):
        """Update the grid size."""
        # Create grid in r
        if self._r_max is None:
            r_max = np.max(np.sqrt(x**2 + y**2))
        else:
            r_max = self._r_max
        if self._r_lim is not None:
            r_max = r_max if r_max <= self._r_lim else self._r_lim
        self._r_max_hist.append(r_max)
        self.dr = r_max / (self.nr - self.nr_border)
        r_max += self.nr_border * self.dr
        self.r_grid = np.linspace(self.dr / 2, r_max - self.dr / 2, self.nr)

        # Create grid in xi
        xi_min_beam = np.min(xi)
        xi_max_beam = np.max(xi)
        self.i_grid = np.where(
            (self.xi_plasma > xi_min_beam - self.dxi * (1 + self.nxi_border))
            & (self.xi_plasma < xi_max_beam + self.dxi * (1 + self.nxi_border))
        )[0]
        self.xi_grid = self.xi_plasma[self.i_grid]
        self.xi_max = self.xi_grid[-1]
        self.xi_min = self.xi_grid[0]
        self.nxi = self.xi_grid.shape[0]

        # Create field arrays.
        self.psi_grid = np.zeros((self.nxi + 4, self.nr + 4))
        self.b_t = np.zeros((self.nxi + 4, self.nr + 4))
        self.e_r = np.zeros((self.nxi + 4, self.nr + 4))
        self.e_z = np.zeros((self.nxi + 4, self.nr + 4))
        self.b_t_bunch = np.zeros((self.nxi + 4, self.nr + 4))
        self.q_bunch = np.zeros((self.nxi + 4, self.nr + 4))

    def _reset_fields(self):
        """Reset value of the fields at the grid."""
        self.psi_grid[:] = 0.0
        self.b_t[:] = 0.0
        self.e_r[:] = 0.0
        self.e_z[:] = 0.0


@njit_serial()
def calculate_fields_on_grid(
    i_grid,
    r_grid,
    s_d,
    psi_grid,
    bt_grid,
    r_hist,
    log_r_hist,
    sum_1_hist,
    sum_2_hist,
    a_0_hist,
    a_i_hist,
    b_i_hist,
):
    """Compute the plasma fields on the grid.

    Compiling this method in numba avoids significant overhead.
    """
    n_points = i_grid.shape[0]
    n_elec = int(r_hist.shape[-1] / 2)
    for i in range(n_points):
        j = i_grid[i]
        psi = psi_grid[i + 2, 2:-2]
        b_theta = bt_grid[i + 2, 2:-2]
        calculate_psi_with_interpolation(
            r_eval=r_grid / s_d,
            r=r_hist[j, :n_elec],
            log_r=log_r_hist[j, :n_elec],
            sum_1_arr=sum_1_hist[j, : n_elec + 1],
            sum_2_arr=sum_2_hist[j, : n_elec + 1],
            psi=psi,
        )
        calculate_psi_with_interpolation(
            r_eval=r_grid / s_d,
            r=r_hist[j, n_elec:],
            log_r=log_r_hist[j, n_elec:],
            sum_1_arr=sum_1_hist[j, n_elec + 1 :],
            sum_2_arr=sum_2_hist[j, n_elec + 1 :],
            psi=psi,
            add=True,
        )
        calculate_b_theta_with_interpolation(
            r_fld=r_grid / s_d,
            a_0=a_0_hist[j],
            a=a_i_hist[j],
            b=b_i_hist[j],
            r=r_hist[j, :n_elec],
            b_theta=b_theta,
        )



from .simple_blowout import SimpleBlowoutWakefield
from .custom_blowout import CustomBlowoutWakefield
from .qs_cold_fluid_1x3p import NonLinearColdFluidWakefield
from .qs_rz_baxevanis import Quasistatic2DWakefield
from .focusing_blowout import FocusingBlowoutField
from .qs_rz_baxevanis_ion import Quasistatic2DWakefieldIon


__all__ = [
    "SimpleBlowoutWakefield",
    "CustomBlowoutWakefield",
    "NonLinearColdFluidWakefield",
    "Quasistatic2DWakefield",
    "FocusingBlowoutField",
    "Quasistatic2DWakefieldIon",
]



"""This module contains the definition of the PlasmaStage class"""

from inspect import signature
from typing import Optional, Union, Callable, List, Literal

import numpy as np
import scipy.constants as ct

import wake_t.physics_models.plasma_wakefields as wf
from wake_t.fields.base import Field
from .field_element import FieldElement
from wake_t.particles.particle_bunch import ParticleBunch


DtBunchType = Union[float, str, List[Union[float, str]]]


wakefield_models = {
    "simple_blowout": wf.SimpleBlowoutWakefield,
    "custom_blowout": wf.CustomBlowoutWakefield,
    "focusing_blowout": wf.FocusingBlowoutField,
    "cold_fluid_1d": wf.NonLinearColdFluidWakefield,
    "quasistatic_2d_legacy": wf.Quasistatic2DWakefield,
    "quasistatic_2d": wf.Quasistatic2DWakefieldIon,  # For backward compatibility
    "quasistatic_2d_ion": wf.Quasistatic2DWakefieldIon,
}


class PlasmaStage(FieldElement):
    """
    Main class for defining a plasma acceleration stage.

    Parameters
    ----------
    length : float
        Length of the plasma stage in m.
    density : float
        Plasma density in units of m^{-3}.
    wakefield_model : str or Field, optional
        Wakefield model to be used. Possible values are ``'blowout'``,
        ``'custom_blowout'``, ``'focusing_blowout'``, ``'cold_fluid_1d'``
        and ``'quasistatic_2d'``. If ``None``, no wakefields will be
        computed.
    bunch_pusher : str, optional
        The pusher used to evolve the particle bunches in time within
        the specified fields. Possible values are ``'rk4'`` (Runge-Kutta
        method of 4th order) or ``'boris'`` (Boris method).
    dt_bunch : float, str or list of float or str, optional
        The time step for evolving the particle bunches. If ``'auto'``, it
        will be automatically set to ``dt = T/(10*2*pi)``, where T is the
        smallest expected betatron period of the bunch along the plasma
        stage. A list of values can also be provided. In this case, the list
        should have the same order as the list of bunches given to the
        ``track`` method.
    auto_dt_bunch : callable, optional
        Function used to determine the adaptive time step for bunches in
        which the time step is set to ``'auto'``. The function should take
        solely a ``ParticleBunch`` as argument.
    push_bunches_before_diags : bool, optional
        Whether to push the bunches before saving them to the diagnostics.
        Since the time step of the diagnostics can be different from that
        of the bunches, it could happen that the bunches appear in the
        diagnostics as they were at the last push, but not at the actual
        time of the diagnostics. Setting this parameter to ``True``
        (default) ensures that an additional push is given to all bunches
        to evolve them to the diagnostics time before saving.
        This additional push will always have a time step smaller than
        the the time step of the bunch, so it has no detrimental impact
        on the accuracy of the simulation. However, it could make
        convergence studies more difficult to interpret,
        since the number of pushes will depend on `n_diags`. Therefore,
        it is exposed as an option so that it can be disabled if needed.
    n_out : int, optional
        Number of times along the stage in which the particle distribution
        should be returned (A list with all output bunches is returned
        after tracking).
    name : str
        Name of the plasma stage. This is only used for displaying the
        progress bar during tracking. By default, ``'Plasma stage'``.
    external_fields : list of Field, optional
        A list of fields to apply to the particle bunches in
        addition to the plasma wakefields.
    **model_params
        Keyword arguments which will be given to the wakefield model. Each
        model requires a different set of parameters. See the documentation
        for each of them for more details.

    See Also
    --------
    wake_t.physics_models.plasma_wakefields.Quasistatic2DWakefield
    wake_t.physics_models.plasma_wakefields.NonLinearColdFluidWakefield

    """

    def __init__(
        self,
        length: float,
        density: Union[float, Callable[[float], float]],
        wakefield_model: Optional[str] = "simple_blowout",
        bunch_pusher: Optional[Literal["boris", "rk4"]] = "boris",
        dt_bunch: Optional[DtBunchType] = "auto",
        auto_dt_bunch: Optional[Callable[[ParticleBunch], float]] = None,
        push_bunches_before_diags: Optional[bool] = True,
        n_out: Optional[int] = 1,
        name: Optional[str] = "Plasma stage",
        external_fields: Optional[List[Field]] = [],
        **model_params,
    ) -> None:
        self.density = self._get_density_profile(density)
        self.wakefield = self._get_wakefield(wakefield_model, model_params)
        self.external_fields = external_fields
        fields = []
        if self.wakefield is not None:
            fields.append(self.wakefield)
        fields.extend(self.external_fields)
        if auto_dt_bunch is not None:
            self.auto_dt_bunch = auto_dt_bunch
        else:
            self.auto_dt_bunch = self._get_optimized_dt
        super().__init__(
            length=length,
            dt_bunch=dt_bunch,
            bunch_pusher=bunch_pusher,
            n_out=n_out,
            name=name,
            fields=fields,
            auto_dt_bunch=self.auto_dt_bunch,
            push_bunches_before_diags=push_bunches_before_diags,
        )

    def _get_density_profile(self, density):
        """Get density profile function"""
        if isinstance(density, float):

            def uniform_density(z, r):
                return np.ones_like(z) * np.ones_like(r) * density

            return uniform_density
        elif callable(density):
            sig = signature(density)
            n_inputs = len(sig.parameters)
            if n_inputs == 2:
                return density
            elif n_inputs == 1:
                # For backward compatibility when only z was supported.
                def density_2d(z, r):
                    return density(z)

                return density_2d
            else:
                raise ValueError(
                    "The density function must take 2 arguments. "
                    "The provided function has {} arguments.".format(n_inputs)
                )
        else:
            raise ValueError("Type {} not supported for density.".format(type(density)))

    def _get_wakefield(self, model, model_params):
        """Initialize and return corresponding wakefield model."""
        if model is None:
            return None
        elif isinstance(model, Field):
            return model
        elif model in wakefield_models:
            return wakefield_models[model](self.density, **model_params)
        else:
            raise ValueError('Wakefield model "{}" not recognized.'.format(model))

    def _get_optimized_dt(self, beam):
        """Get tracking time step."""
        # Get minimum gamma in the bunch (assumes px,py << pz).
        min_gamma = np.sqrt(np.min(beam.pz) ** 2 + 1)
        # calculate maximum focusing along stage.
        z = np.linspace(0, self.length, 100)
        n_p = self.density(z, 0.0)
        q_over_m = beam.q_species / beam.m_species
        w_p = np.sqrt(max(n_p) * ct.e**2 / (ct.m_e * ct.epsilon_0))
        max_kx = (ct.m_e / (2 * ct.e * ct.c)) * w_p**2
        w_x = np.sqrt(np.abs(q_over_m * ct.c * max_kx / min_gamma))
        period_x = 1 / w_x
        dt = 0.1 * period_x
        return dt



"""
Contains the method to compute the wakefield potential and its derivatives
according to the paper by P. Baxevanis and G. Stupakov.

"""

import numpy as np

from wake_t.utilities.numba import njit_serial


@njit_serial()
def calculate_psi_and_derivatives_at_particles(
    r, pr, q, idx, r_max, dr_p, pc, psi_pp, dr_psi_pp, dxi_psi_pp
):
    """
    Calculate the wakefield potential and its derivatives at the position
    of the plasma particles. This is done by using Eqs. (29) - (32) in
    the paper by P. Baxevanis and G. Stupakov.

    As indicated in the original paper, the value of the fields at the
    discontinuities (at the exact radial position of the plasma particles)
    is calculated as the average between the two neighboring values.

    Parameters
    ----------
    r, pr, q : array
        Arrays containing the radial position, momentum and charge of the
        plasma particles.
    idx : ndarray
        Array containing the (radially) sorted indices of the plasma particles.
    r_max : float
        Maximum radial extent of the plasma column.
    dr_p : float
        Initial spacing between plasma macroparticles. Corresponds also the
        width of the plasma sheet represented by the macroparticle.
    psi_pp, dr_psi_pp, dxi_psi_pp : ndarray
        Arrays where the value of the wakefield potential and its derivatives
        at the location of the plasma particles will be stored.

    """
    # Initialize arrays.
    n_part = r.shape[0]

    # Initialize value of sums.
    sum_1 = 0.0
    sum_2 = 0.0
    sum_3 = 0.0

    # Calculate psi and dr_psi.
    # Their value at the position of each plasma particle is calculated
    # by doing a linear interpolation between two values at the left and
    # right of the particle. The left point is the middle position between the
    # particle and its closest left neighbor, and the same for the right.
    for i_sort in range(n_part):
        i = idx[i_sort]
        r_i = r[i]
        q_i = q[i]

        # Calculate new sums.
        sum_1_new = sum_1 + q_i
        sum_2_new = sum_2 + q_i * np.log(r_i)

        # If this is not the first particle, calculate the left point (r_left)
        # and the field values there (psi_left and dr_psi_left) as usual.
        if i_sort > 0:
            r_im1 = r[idx[i_sort - 1]]
            r_left = (r_im1 + r_i) / 2
            psi_left = delta_psi_eq(r_left, sum_1, sum_2, r_max, pc)
            dr_psi_left = dr_psi_eq(r_left, sum_1, r_max, pc)
        # Otherwise, take r=0 as the location of the left point.
        else:
            r_left = 0.0
            psi_left = 0.0
            dr_psi_left = 0.0

        # If this is not the last particle, calculate the r_right as
        # middle point.
        if i_sort < n_part - 1:
            r_ip1 = r[idx[i_sort + 1]]
            r_right = (r_i + r_ip1) / 2
        # Otherwise, since the particle represents a charge sheet of width
        # dr_p, take the right point as r_i + dr_p/2.
        else:
            r_right = r_i + dr_p / 2
        # Calculate field values ar r_right.
        psi_right = delta_psi_eq(r_right, sum_1_new, sum_2_new, r_max, pc)
        dr_psi_right = dr_psi_eq(r_right, sum_1_new, r_max, pc)

        # Interpolate psi.
        b_1 = (psi_right - psi_left) / (r_right - r_left)
        a_1 = psi_left - b_1 * r_left
        psi_pp[i] = a_1 + b_1 * r_i

        # Interpolate dr_psi.
        b_2 = (dr_psi_right - dr_psi_left) / (r_right - r_left)
        a_2 = dr_psi_left - b_2 * r_left
        dr_psi_pp[i] = a_2 + b_2 * r_i

        # Update value of sums.
        sum_1 = sum_1_new
        sum_2 = sum_2_new

    # Boundary condition for psi (Force potential to be zero either at the
    # plasma edge or after the last particle, whichever is further away).
    r_furthest = max(r_right, r_max)
    psi_pp -= delta_psi_eq(r_furthest, sum_1, sum_2, r_max, pc)

    # In theory, psi cannot be smaller than -1. However, it has been observed
    # than in very strong blowouts, near the peak, values below -1 can appear
    # in this numerical method. In addition, values very close to -1 will lead
    # to particles with gamma >> 10, which will also lead to problems.
    # This condition here makes sure that this does not happen, improving
    # the stability of the solver.
    for i in range(n_part):
        # Should only happen close to the peak of very strong blowouts.
        if psi_pp[i] < -0.90:
            psi_pp[i] = -0.90

    # Calculate dxi_psi (also by interpolation).
    for i_sort in range(n_part):
        i = idx[i_sort]
        r_i = r[i]
        pr_i = pr[i]
        q_i = q[i]
        psi_i = psi_pp[i]

        sum_3_new = sum_3 + (q_i * pr_i) / (r_i * (1 + psi_i))

        # Check if it is the first particle.
        if i_sort > 0:
            r_im1 = r[idx[i_sort - 1]]
            r_left = (r_im1 + r_i) / 2
            dxi_psi_left = -sum_3
        else:
            r_left = 0.0
            dxi_psi_left = 0.0

        # Check if it is the last particle.
        if i_sort < n_part - 1:
            r_ip1 = r[idx[i_sort + 1]]
            r_right = (r_i + r_ip1) / 2
        else:
            r_right = r_i + dr_p / 2
        dxi_psi_right = -sum_3_new

        # Do interpolation.
        b = (dxi_psi_right - dxi_psi_left) / (r_right - r_left)
        a = dxi_psi_left - b * r_left
        dxi_psi_pp[i] = a + b * r_i
        sum_3 = sum_3_new

    # Apply longitudinal derivative of the boundary conditions of psi.
    if r_right <= r_max:
        dxi_psi_pp += sum_3
    else:
        dxi_psi_pp += sum_3 - (
            (sum_1 - r_max**2 / 2 - pc * r_max / 4) * pr_i / (r_right * (1 + psi_i))
        )

    # Again, near the peak of a strong blowout, very large and unphysical
    # values could appear. This condition makes sure a threshold us not
    # exceeded.
    for i in range(n_part):
        if dxi_psi_pp[i] > 3.0:
            dxi_psi_pp[i] = 3.0
        if dxi_psi_pp[i] < -3.0:
            dxi_psi_pp[i] = -3.0


@njit_serial()
def calculate_psi(r_fld, r, q, idx, r_max, pc, psi, k):
    """
    Calculate the wakefield potential at the radial
    positions specified in r_fld. This is done by using Eq. (29) in
    the paper by P. Baxevanis and G. Stupakov.

    Parameters
    ----------
    r_fld : array
        Array containing the radial positions where psi should be calculated.
    r, q : array
        Arrays containing the radial position, and charge of the
        plasma particles.
    idx : ndarray
        Array containing the (radially) sorted indices of the plasma particles.
    r_max : float
        Maximum radial extent of the plasma column.
    pc : float
        The parabolic density profile coefficient.
    psi : ndarray
        Array where the values of the wakefield potential will be stored.
    k : int
        Index that determines the slice of psi where the values will
        be filled in (the index is k+2 due to the guard cells in the array).

    """
    # Initialize arrays with values of psi and sums at plasma particles.
    n_part = r.shape[0]
    sum_1_arr = np.zeros(n_part)
    sum_2_arr = np.zeros(n_part)
    sum_1 = 0.0
    sum_2 = 0.0

    # Calculate sum_1, sum_2 and psi_part.
    for i_sort in range(n_part):
        i = idx[i_sort]
        r_i = r[i]
        q_i = q[i]

        sum_1 += q_i
        sum_2 += q_i * np.log(r_i)
        sum_1_arr[i] = sum_1
        sum_2_arr[i] = sum_2
    r_N = r_i

    # Initialize array for psi at r_fld locations.
    n_points = r_fld.shape[0]
    psi_slice = psi[k + 2]

    # Calculate fields at r_fld.
    i_last = 0
    for j in range(n_points):
        r_j = r_fld[j]
        # Get index of last plasma particle with r_i < r_j, continuing from
        # last particle found in previous iteration.
        for i_sort in range(i_last, n_part):
            i = idx[i_sort]
            r_i = r[i]
            i_last = i_sort
            if r_i >= r_j:
                i_last -= 1
                break
        # Calculate fields at r_j.
        if i_last == -1:
            sum_1_j = 0.0
            sum_2_j = 0.0
            i_last = 0
        else:
            i = idx[i_last]
            sum_1_j = sum_1_arr[i]
            sum_2_j = sum_2_arr[i]
        psi_slice[2 + j] = delta_psi_eq(r_j, sum_1_j, sum_2_j, r_max, pc)

    # Apply boundary conditions.
    r_furthest = max(r_N, r_max)
    psi_slice -= delta_psi_eq(r_furthest, sum_1, sum_2, r_max, pc)


@njit_serial()
def calculate_psi_and_derivatives(r_fld, r, pr, q):
    """
    Calculate the wakefield potential and its derivatives at the radial
    positions specified in r_fld. This is done by using Eqs. (29) - (32) in
    the paper by P. Baxevanis and G. Stupakov.

    Parameters
    ----------
    r_fld : array
        Array containing the radial positions where psi should be calculated.
    r, pr, q : array
        Arrays containing the radial position, momentum and charge of the
        plasma particles.

    """
    # Initialize arrays with values of psi and sums at plasma particles.
    n_part = r.shape[0]
    psi_part = np.zeros(n_part)
    sum_1_arr = np.zeros(n_part)
    sum_2_arr = np.zeros(n_part)
    sum_3_arr = np.zeros(n_part)
    sum_1 = 0.0
    sum_2 = 0.0
    sum_3 = 0.0

    # Calculate sum_1, sum_2 and psi_part.
    idx = np.argsort(r)
    for i_sort in range(n_part):
        i = idx[i_sort]
        r_i = r[i]
        pr_i = pr[i]
        q_i = q[i]

        sum_1 += q_i
        sum_2 += q_i * np.log(r_i)
        sum_1_arr[i] = sum_1
        sum_2_arr[i] = sum_2
        psi_part[i] = sum_1 * np.log(r_i) - sum_2 - 0.25 * r_i**2
    r_N = r[-1]
    psi_part += -(sum_1 * np.log(r_N) - sum_2 - 0.25 * r_N**2)

    # Calculate sum_3.
    for i_sort in range(n_part):
        i = idx[i_sort]
        r_i = r[i]
        pr_i = pr[i]
        q_i = q[i]
        psi_i = psi_part[i]

        sum_3 += (q_i * pr_i) / (r_i * (1 + psi_i))
        sum_3_arr[i] = sum_3

    # Initialize arrays for psi and derivatives at r_fld locations.
    n_points = r_fld.shape[0]
    psi = np.zeros(n_points)
    dr_psi = np.zeros(n_points)
    dxi_psi = np.zeros(n_points)

    # Calculate fields at r_fld.
    i_last = 0
    for j in range(n_points):
        r_j = r_fld[j]
        # Get index of last plasma particle with r_i < r_j.
        for i_sort in range(n_part):
            i = idx[i_sort]
            r_i = r[i]
            i_last = i_sort
            if r_i >= r_j:
                i_last -= 1
                break
        # Calculate fields at r_j.
        if i_last == -1:
            psi[j] = -0.25 * r_j**2
            dr_psi[j] = -0.5 * r_j
            dxi_psi[j] = 0.0
        else:
            i_p = idx[i_last]
            psi[j] = sum_1_arr[i_p] * np.log(r_j) - sum_2_arr[i_p] - 0.25 * r_j**2
            dr_psi[j] = sum_1_arr[i_p] / r_j - 0.5 * r_j
            dxi_psi[j] = -sum_3_arr[i_p]
    psi = psi - (sum_1 * np.log(r_N) - sum_2 - 0.25 * r_N**2)
    dxi_psi = dxi_psi + sum_3
    return psi, dr_psi, dxi_psi


@njit_serial()
def delta_psi_eq(r, sum_1, sum_2, r_max, pc):
    """Adapted equation (29) from original paper."""
    delta_psi_elec = sum_1 * np.log(r) - sum_2
    if r <= r_max:
        delta_psi_ion = 0.25 * r**2 + pc * r**4 / 16
    else:
        delta_psi_ion = (
            0.25 * r_max**2
            + pc * r_max**4 / 16
            + (0.5 * r_max**2 + 0.25 * pc * r_max**4) * (np.log(r) - np.log(r_max))
        )
    return delta_psi_elec - delta_psi_ion


@njit_serial()
def dr_psi_eq(r, sum_1, r_max, pc):
    """Adapted equation (31) from original paper."""
    dr_psi_elec = sum_1 / r
    if r <= r_max:
        dr_psi_ion = 0.5 * r + 0.25 * pc * r**3
    else:
        dr_psi_ion = (0.5 * r_max**2 + 0.25 * pc * r_max**4) / r
    return dr_psi_elec - dr_psi_ion



import os

import pytest
from numpy.testing import assert_array_almost_equal

from wake_t.utilities.bunch_generation import (
    get_from_file,
    get_gaussian_bunch_from_size,
)
from wake_t.utilities.bunch_saving import save_bunch_to_file
from wake_t.diagnostics import OpenPMDDiagnostics


tests_output_folder = "./tests_output"


def test_bunch_read_save():
    """
    Check that saving/reading a particle bunch to/from a file does not
    have any effect on the data.

    All supported data formats (openpmd, astra and csrtrack) are tested.
    """
    output_folder = os.path.join(tests_output_folder, "bunch_read_save")
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Beam parameters.
    emitt_nx = emitt_ny = 1e-6  # m
    s_x = s_y = 3e-6  # m
    s_t = 3.0  # fs
    gamma_avg = 100 / 0.511
    gamma_spread = 1.0  # %
    q_bunch = 30  # pC
    xi_avg = 0.0  # m
    n_part = 1e4

    # Create particle bunch.
    bunch = get_gaussian_bunch_from_size(
        emitt_nx,
        emitt_ny,
        s_x,
        s_y,
        gamma_avg,
        gamma_spread,
        s_t,
        xi_avg,
        q_bunch,
        n_part,
        name="elec_bunch",
    )

    # Formats to test.
    data_formats = ["astra", "csrtrack", "openpmd"]
    file_formats = ["txt", "fmt1", "h5"]

    for data_format, file_format in zip(data_formats, file_formats):
        # Save bunch.
        save_bunch_to_file(
            bunch,
            data_format,
            os.path.join(output_folder, "bunch_{}.{}".format(data_format, file_format)),
        )
        # Read saved bunch.
        file_path = os.path.join(
            output_folder, "bunch_{}.{}".format(data_format, file_format)
        )
        bunch_saved = get_from_file(file_path, data_format)

        # For astra and csrtrack, remove the generated reference particle.
        if data_format in ["astra", "csrtrack"]:
            bunch_saved.x = bunch_saved.x[1:]
            bunch_saved.y = bunch_saved.y[1:]
            bunch_saved.xi = bunch_saved.xi[1:]
            bunch_saved.px = bunch_saved.px[1:]
            bunch_saved.py = bunch_saved.py[1:]
            bunch_saved.pz = bunch_saved.pz[1:]
            bunch_saved.w = bunch_saved.w[1:]

        # Check saved bunch is the same as original.
        assert_array_almost_equal(bunch_saved.x, bunch.x)
        assert_array_almost_equal(bunch_saved.y, bunch.y)
        assert_array_almost_equal(bunch_saved.xi, bunch.xi)
        assert_array_almost_equal(bunch_saved.px, bunch.px)
        assert_array_almost_equal(bunch_saved.py, bunch.py)
        assert_array_almost_equal(bunch_saved.pz, bunch.pz)
        assert_array_almost_equal(bunch_saved.q, bunch.q)


def test_openpmd_reading():
    """
    Check that the `get_from_file` method behaves as expected when reading
    openPMD data.
    """
    output_folder = os.path.join(tests_output_folder, "bunch_openpmd_read")
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Beam parameters.
    emitt_nx = emitt_ny = 1e-6  # m
    s_x = s_y = 3e-6  # m
    s_t = 3.0  # fs
    gamma_avg = 100 / 0.511
    gamma_spread = 1.0  # %
    q_bunch = 30  # pC
    xi_avg = 0.0  # m
    n_part = 1e4

    # Create particle bunch.
    bunch_1 = get_gaussian_bunch_from_size(
        emitt_nx,
        emitt_ny,
        s_x,
        s_y,
        gamma_avg,
        gamma_spread,
        s_t,
        xi_avg,
        q_bunch,
        n_part,
        name="elec_bunch",
    )
    bunch_2 = get_gaussian_bunch_from_size(
        emitt_nx,
        emitt_ny,
        s_x,
        s_y,
        gamma_avg,
        gamma_spread,
        s_t,
        xi_avg,
        q_bunch,
        n_part,
        name="elec_bunch_2",
    )

    # Create diagnostics in output folder
    diags = OpenPMDDiagnostics(write_dir=output_folder)

    # Test 1.
    # -------

    # Write diagnostics without any species.
    diags.write_diagnostics(time=0.0, dt=0.0)
    # Trying to read a species from this file should fail.
    with pytest.raises(ValueError):
        file_path = os.path.join(output_folder, "hdf5", "data00000000.h5")
        bunch_saved = get_from_file(file_path, "openpmd")

    # Test 2.
    # -------

    # Write diagnostics with one species.
    diags.write_diagnostics(time=0.0, dt=0.0, species_list=[bunch_1])
    # Reading this file should raise no exception even if no species name
    # is specified.
    file_path = os.path.join(output_folder, "hdf5", "data00000001.h5")
    bunch_saved = get_from_file(file_path, "openpmd")
    # Since the `name` parameter has not been specified, it must be equal to
    # the original name of the openPMD species.
    assert bunch_saved.name == "elec_bunch"
    # If a new custom name is given, use this name.
    custom_name = "bunch"
    bunch_saved = get_from_file(file_path, "openpmd", name=custom_name)
    assert bunch_saved.name == custom_name

    # Test 3.
    # -------

    # Write diagnostics with two species.
    diags.write_diagnostics(time=0.0, dt=0.0, species_list=[bunch_1, bunch_2])
    # Reading a species without specifying a species name should fail.
    with pytest.raises(ValueError):
        file_path = os.path.join(output_folder, "hdf5", "data00000002.h5")
        bunch_saved = get_from_file(file_path, "openpmd")
    # Reading each species indicating the correct species name should not fail
    bunch_saved = get_from_file(file_path, "openpmd", species_name="elec_bunch")
    bunch_saved = get_from_file(file_path, "openpmd", species_name="elec_bunch_2")


if __name__ == "__main__":
    test_bunch_read_save()
    test_openpmd_reading()



from typing import Optional, Callable

import numpy as np
import scipy.constants as ct
import aptools.plasma_accel.general_equations as ge

from wake_t.particles.deposition import deposit_3d_distribution
from wake_t.particles.interpolation import gather_field_cyl_linear
from wake_t.fields.rz_wakefield import RZWakefield
from wake_t.physics_models.laser.laser_pulse import LaserPulse


class NonLinearColdFluidWakefield(RZWakefield):
    """
    This class computes the plasma wakefields using a nonlinear cold fluid
    theory in one spatial dimension with a three-component fluid momentum, as
    described in [1]_. This implies that only longitudinal plasma waves are
    modeled, i.e., it assumes infinitely broad laser pulses and particle beams.

    This 1D model is in Wake-T extended to 2D in r-z geometry by computing the
    1D plasma response at each radial slice.

    Given the assumptions of the model, it is only accurate for broad drivers
    where the radial plasma waves can be neglected. For a laser driver, it is
    typically accurate up to :math:`a_0 \\lesssim 1`. For electron beams it has
    not been yet fully tested, but given the typically narrow width of the
    beams, only very low charges can be accurately modeled.

    For a much more general model, see :class:`Quasistatic2DWakefield`.

    Parameters
    ----------
    density_function : callable
        Function that returns the density value at the given position z.
        This parameter is given by the `PlasmaStage` and does not need
        to be specified by the user.
    r_max : float
        Maximum radial position up to which plasma wakefield will be
        calculated. Required only if mode='cold_fluid_1d'.
    xi_min : float
        Minimum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    xi_max : float
        Maximum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    n_r : int
        Number of grid elements along r to calculate the wakefields.
    n_xi : int
        Number of grid elements along xi to calculate the wakefields.
    dz_fields : float, optional
        Determines how often the plasma wakefields should be updated. If
        dz_fields=0 (default value), the wakefields are calculated at every
        step of the Runge-Kutta solver for the beam particle evolution
        (most expensive option). If specified, the wakefields are only
        updated in steps determined by dz_fields. For example, if
        dz_fields=10e-6, the plasma wakefields are only updated every time
        the simulation window advances by 10 micron. By default, if not
        specified, the value of `dz_fields` will be `xi_max-xi_min`, i.e.,
        the length the simulation box.
    beam_wakefields : bool, optional
        Whether to take into account beam-driven wakefields (False by
        default). This should be set to True for any beam-driven case or
        in order to take into account the beam-loading of the witness in
        a laser-driven case.
    p_shape : str, optional
        Particle shape to be used for the beam charge deposition. Possible
        values are 'linear' or 'cubic'.
    laser : LaserPulse, optional
        Laser driver of the plasma stage.
    laser_evolution : bool, optional
        If True (default), the laser pulse is evolved
        using a laser envelope model. If False, the pulse envelope stays
        unchanged throughout the computation.
    laser_envelope_substeps : int, optional
        Number of substeps of the laser envelope solver per `dz_fields`.
        The time step of the envelope solver is therefore
        `dz_fields / c / laser_envelope_substeps`.
    laser_envelope_nxi, laser_envelope_nr : int, optional
        If given, the laser envelope will run in a grid of size
        (`laser_envelope_nxi`, `laser_envelope_nr`) instead
        of (`n_xi`, `n_r`). This allows the laser to run in a finer (or
        coarser) grid than the plasma wake. It is not necessary to specify
        both parameters. If one of them is not given, the resolution of
        the plasma grid with be used for that direction.
    laser_envelope_use_phase : bool, optional
        Determines whether to take into account the terms related to the
        longitudinal derivative of the complex phase in the envelope
        solver.

    See Also
    --------
    Quasistatic2DWakefield

    References
    ----------
    .. [1] T. Mehrling, "Theoretical and numerical studies on the transport of
       transverse beam quality in plasma-based accelerators," PhD thesis
       (2014), http://dx.doi.org/10.3204/DESY-THESIS-2014-040

    """

    def __init__(
        self,
        density_function: Callable[[float], float],
        r_max: float,
        xi_min: float,
        xi_max: float,
        n_r: int,
        n_xi: int,
        dz_fields: Optional[float] = None,
        beam_wakefields: Optional[bool] = False,
        p_shape: Optional[str] = "linear",
        laser: Optional[LaserPulse] = None,
        laser_evolution: Optional[bool] = True,
        laser_envelope_substeps: Optional[int] = 1,
        laser_envelope_nxi: Optional[int] = None,
        laser_envelope_nr: Optional[int] = None,
        laser_envelope_use_phase: Optional[bool] = True,
    ) -> None:
        self.beam_wakefields = beam_wakefields
        self.p_shape = p_shape
        super().__init__(
            density_function=density_function,
            r_max=r_max,
            xi_min=xi_min,
            xi_max=xi_max,
            n_r=n_r,
            n_xi=n_xi,
            dz_fields=dz_fields,
            laser=laser,
            laser_evolution=laser_evolution,
            laser_envelope_substeps=laser_envelope_substeps,
            laser_envelope_nxi=laser_envelope_nxi,
            laser_envelope_nr=laser_envelope_nr,
            laser_envelope_use_phase=laser_envelope_use_phase,
            model_name="cold_fluid_1d",
        )

    def __wakefield_ode_system(self, u_1, u_2, laser_a0, n_beam):
        if self.beam_wakefields:
            return np.array(
                [u_2, (1 + laser_a0**2) / (2 * (1 + u_1) ** 2) - n_beam - 1 / 2]
            )
        else:
            return np.array([u_2, (1 + laser_a0**2) / (2 * (1 + u_1) ** 2) - 1 / 2])

    def _calculate_wakefield(self, bunches):
        # Get laser envelope
        if self.laser is not None:
            a_env = np.abs(self.laser.get_envelope())
            # If linearly polarized, divide by sqrt(2) so that the
            # ponderomotive force on the plasma particles is correct.
            if self.laser.polarization == "linear":
                a_env /= np.sqrt(2)
        else:
            a_env = np.zeros((self.n_xi, self.n_r))

        # Calculate and allocate laser quantities, including guard cells.
        a_rz = np.zeros((self.n_xi + 4, self.n_r + 4))
        a_rz[2:-2, 2:-2] = a_env

        s_d = ge.plasma_skin_depth(self.n_p * 1e-6)
        dz = self.dxi / s_d
        dr = self.dr / s_d
        r_fld = self.r_fld / s_d

        # Get charge distribution and remove guard cells.
        beam_hist = np.zeros((self.n_xi + 4, self.n_r + 4))
        for bunch in bunches:
            x = bunch.x
            y = bunch.y
            xi = bunch.xi
            w = bunch.w * (bunch.q_species / ct.e)
            deposit_3d_distribution(
                xi / s_d,
                x / s_d,
                y / s_d,
                w,
                self.xi_min / s_d,
                r_fld[0],
                self.n_xi,
                self.n_r,
                dz,
                dr,
                beam_hist,
                p_shape=self.p_shape,
                use_ruyten=True,
            )
        beam_hist = beam_hist[2:-2, 2:-2]

        n = np.arange(self.n_r)
        disc_area = np.pi * dr**2 * (1 + 2 * n)
        beam_hist *= 1 / (disc_area * dz * self.n_p) / s_d**3
        n_iter = self.n_xi - 1
        u_1 = np.zeros((n_iter + 1, len(r_fld)))
        u_2 = np.zeros((n_iter + 1, len(r_fld)))
        z_fld = self.xi_fld / s_d

        for i in np.arange(n_iter):
            z_i = z_fld[-1 - i]
            # get laser a0 at z, z+dz/2 and z+dz
            if self.laser is not None:
                x = r_fld
                y = np.zeros_like(r_fld)
                z = np.full_like(r_fld, z_i)
                a0_0 = gather_field_cyl_linear(
                    a_rz,
                    self.xi_min / s_d,
                    self.xi_max / s_d,
                    r_fld[0],
                    r_fld[-1],
                    dz,
                    dr,
                    x,
                    y,
                    z,
                )
                a0_1 = gather_field_cyl_linear(
                    a_rz,
                    self.xi_min / s_d,
                    self.xi_max / s_d,
                    r_fld[0],
                    r_fld[-1],
                    dz,
                    dr,
                    x,
                    y,
                    z - dz / 2,
                )
                a0_2 = gather_field_cyl_linear(
                    a_rz,
                    self.xi_min / s_d,
                    self.xi_max / s_d,
                    r_fld[0],
                    r_fld[-1],
                    dz,
                    dr,
                    x,
                    y,
                    z - dz,
                )
            else:
                a0_0 = np.zeros(r_fld.shape[0])
                a0_1 = np.zeros(r_fld.shape[0])
                a0_2 = np.zeros(r_fld.shape[0])
            # perform runge-kutta
            A = dz * self.__wakefield_ode_system(
                u_1[-1 - i], u_2[-1 - i], a0_0, beam_hist[-i - 1]
            )
            B = dz * self.__wakefield_ode_system(
                u_1[-1 - i] + A[0] / 2, u_2[-1 - i] + A[1] / 2, a0_1, beam_hist[-i - 1]
            )
            C = dz * self.__wakefield_ode_system(
                u_1[-1 - i] + B[0] / 2, u_2[-1 - i] + B[1] / 2, a0_1, beam_hist[-i - 1]
            )
            D = dz * self.__wakefield_ode_system(
                u_1[-1 - i] + C[0], u_2[-1 - i] + C[1], a0_2, beam_hist[-i - 1]
            )
            u_1[-2 - i] = u_1[-1 - i] + 1 / 6 * (A[0] + 2 * B[0] + 2 * C[0] + D[0])
            u_2[-2 - i] = u_2[-1 - i] + 1 / 6 * (A[1] + 2 * B[1] + 2 * C[1] + D[1])
        E_z = -np.gradient(u_1, dz, axis=0, edge_order=2)
        W_r = -np.gradient(u_1, dr, axis=1, edge_order=2)
        E_0 = ge.plasma_cold_non_relativisct_wave_breaking_field(self.n_p * 1e-6)

        # Calculate rho and chi.
        gamma_fl = (1 + a_env**2 + (1 + u_1) ** 2) / (2 * (1 + u_1))
        rho_fl = gamma_fl / (1 + u_1)
        self.rho[2:-2, 2:-2] = rho_fl
        self.chi[2:-2, 2:-2] = rho_fl / gamma_fl

        # Calculate B_theta and E_r.
        u_z = (1 + a_env**2 - (1 + u_1) ** 2) / (2 * (1 + u_1))
        dE_z = np.gradient(E_z, dz, axis=0, edge_order=2)
        v_z = u_z / gamma_fl
        nv_z = rho_fl * v_z
        integrand = (dE_z - nv_z + beam_hist) * r_fld
        subs = integrand / 2
        B_theta = (np.cumsum(integrand, axis=1) - subs) * dr / np.abs(r_fld)
        E_r = W_r + B_theta

        # Store fields.
        self.b_t[2:-2, 2:-2] = B_theta * E_0 / ct.c
        self.e_r[2:-2, 2:-2] = E_r * E_0
        self.e_z[2:-2, 2:-2] = E_z * E_0



"""Contains methods for interpolating fields to a different grid."""

import math
import numpy as np

from wake_t.utilities.numba import njit_serial


@njit_serial(fastmath=True)
def interpolate_rz_field(
    fld: np.ndarray,
    z_min: float,
    r_min: float,
    dz: float,
    dr: float,
    z_new: np.ndarray,
    r_new: np.ndarray,
    fld_new: np.ndarray,
) -> None:
    """
    Interpolate a field in r-z geometry to a new grid using linear
    interpolation.

    Parameters
    ----------
    fld : 2darray
        The original field to interpolate.
    z_min : float
        Position of the first field value along `z`.
    r_min : float
        Position of the first field value along `r`.
    dz : float
        Grid spacing in `z` of the original field.
    dr : float
        Grid spacing in `r` of the original field.
    z_new : 1darray
        Location of the grid nodes in `z` of the new, interpolated field.
    r_new : 1darray
        Location of the grid nodes in `r` of the new, interpolated field.
    fld_new : 2darray
        Array where to store the new, interpolated field. The shape of this
        array should be (`nz_new`, `nr_new`), here `nz_new` and `nr_new` are
        the length of `z_new` and `r_new`, respectively.
    """
    # Get size of original array.
    nz, nr = fld.shape

    # Iterate in z.
    for i in range(z_new.shape[0]):
        # z value at which to evaluate the array by interpolation.
        z_i = z_new[i]

        # Position of z_i in cell units.
        z_i_cell = (z_i - z_min) / dz

        # Lower and upper cells in z.
        iz_lower = int(math.floor(z_i_cell))
        iz_upper = iz_lower + 1

        # Linear interpolation coefficients in z.
        dz_u = iz_upper - z_i_cell
        dz_l = z_i_cell - iz_lower

        # Make sure iz_upper does not go beyond boundaries.
        iz_upper = min(iz_upper, nz - 1)

        # Iterate in r.
        for j in range(r_new.shape[0]):
            # r value at which to evaluate the array by interpolation.
            r_j = r_new[j]

            # Position of r_i in cell units.
            r_j_cell = (r_j - r_min) / dr

            # Lower and upper cells in r.
            jr_lower = max(int(math.floor(r_j_cell)), 0)
            jr_upper = jr_lower + 1

            # Linear interpolation coefficients in r.
            dr_u = jr_upper - r_j_cell
            dr_l = 1 - dr_u

            # Make sure jz_upper does not go beyond boundaries.
            jr_upper = min(jr_lower + 1, nr - 1)

            # Get field value at each bounding cell.
            fld_ll = fld[iz_lower, jr_lower]
            fld_lu = fld[iz_lower, jr_upper]
            fld_ul = fld[iz_upper, jr_lower]
            fld_uu = fld[iz_upper, jr_upper]

            # If evaluating beyond the last value in r, extrapolate linearly.
            if jr_lower == jr_upper:
                fld_lu += fld_lu - fld[iz_lower, jr_upper - 1]
                fld_uu += fld_uu - fld[iz_upper, jr_upper - 1]

            # Interpolate in z.
            fld_z_1 = dz_u * fld_ll + dz_l * fld_ul
            fld_z_2 = dz_u * fld_lu + dz_l * fld_uu

            # Interpolate in r.
            fld_new[i, j] = dr_u * fld_z_1 + dr_l * fld_z_2



from pytest import approx
import numpy as np
import matplotlib.pyplot as plt
from aptools.plotting.quick_diagnostics import slice_analysis

from wake_t import PlasmaStage, GaussianPulse
from wake_t.utilities.bunch_generation import get_matched_bunch
from wake_t.diagnostics import analyze_bunch, analyze_bunch_list


def test_custom_blowout_wakefield(make_plots=False):
    """Check that the `'custom_blowout'` wakefield model works as expected.

    A matched bunch with high emittance is propagated through a 3 cm plasma
    stage. Due to the high emittance and Ez slope, the beam develops not only
    a large chirp but also an uncorrelated energy spread. The test checks that
    the final projected energy spread is always the same.
    """
    # Set numpy random seed to get reproducible results
    np.random.seed(1)

    # Create laser driver.
    laser = GaussianPulse(100e-6, l_0=800e-9, w_0=50e-6, a_0=3, tau=30e-15, z_foc=0.0)

    # Create bunch (matched to a blowout at a density of 10^{23} m^{-3}).
    en = 10e-6  # m
    ene = 200  # units of beta*gamma
    ene_sp = 0.3  # %
    xi_c = laser.xi_c - 55e-6  # m
    s_t = 10  # fs
    q_tot = 100  # pC
    n_part = 3e4
    k_x = 1e6
    bunch = get_matched_bunch(en, en, ene, ene_sp, s_t, xi_c, q_tot, n_part, k_x=k_x)

    # Create plasma stage.
    plasma = PlasmaStage(
        3e-2,
        1e23,
        laser=laser,
        wakefield_model="custom_blowout",
        lon_field=-10e9,
        lon_field_slope=1e15,
        foc_strength=k_x,
        xi_fields=xi_c,
        n_out=50,
        bunch_pusher="boris",
    )

    # Do tracking.
    bunch_list = plasma.track(bunch)

    bunch_params = analyze_bunch(bunch)
    rel_ene_sp = bunch_params["rel_ene_spread"]
    assert approx(rel_ene_sp, rel=1e-10) == 0.21192494144893628

    if make_plots:
        # Analyze bunch evolution.
        params_evolution = analyze_bunch_list(bunch_list)

        # Quick plot of results.
        z = params_evolution["prop_dist"] * 1e2
        plt.figure()
        plt.subplot(411)
        plt.plot(z, params_evolution["beta_x"] * 1e3)
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\beta_x$ [mm]")
        plt.subplot(412)
        plt.plot(z, params_evolution["emitt_x"] * 1e6)
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\epsilon_{nx}$ [$\\mu$m]")
        plt.subplot(413)
        plt.plot(z, params_evolution["rel_ene_spread"] * 100)
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\frac{\\Delta \\gamma}{\\gamma}$ [%]")
        plt.subplot(414)
        plt.plot(z, params_evolution["avg_ene"])
        plt.xlabel("z [mm]")
        plt.ylabel("$\\gamma$")
        plt.tight_layout()
        fig_2 = plt.figure()
        slice_analysis(
            bunch.x, bunch.y, bunch.xi, bunch.px, bunch.py, bunch.pz, bunch.q, fig=fig_2
        )
        plt.show()


if __name__ == "__main__":
    test_custom_blowout_wakefield(make_plots=True)



"""Contains the Runge-Kutta pusher of order 4"""

import math
import scipy.constants as ct
from numba import prange

from wake_t.utilities.numba import njit_parallel
from wake_t.fields.gather import gather_fields


def apply_rk4_pusher(bunch, fields, t, dt):
    """Evolve a particle bunch using the RK4 pusher.

    For a variable evolving as:

        dx/dt = f(t, x) ,

    the pusher calculates its value at t = t_n + dt as:

        x += dt * (k1 + 2*k_2 + 2*k_3 + k4) / 6 ,

    where the k_i coefficients are given by:

        k_1 = f(t_n, x_n) ,
        k_2 = f(t_n + dt/2, x_n + dt*k_1/2) ,
        k_3 = f(t_n + dt/2, x_n + dt*k_2/2) ,
        k_4 = f(t_n + dt, x_n + dt*k_3) .

    In order to be more efficient, only a single 1D k_i array per variable is
    allocated. Once k_i is calculated, its contribution to the push is added
    (i.e., x_push += k_i * fac, where fac can be 1/6 or 2/6). The same
    array can then be overwritten with the values of k_i+1, whose contribution
    is also added to x_push. Once the contribution of all four k_i has been
    added, the push is applied (i.e., x += dt * x_push).

    Parameters
    ----------
    bunch : ParticleBunch
        The particle bunch to be evolved.
    fields : list
        List of fields within which the particle bunch will be evolved.
    t : float
        The current time.
    dt : float
        Time step by which to push the particles.
    """
    # Get the necessary preallocated arrays.
    (
        x,
        y,
        xi,
        px,
        py,
        pz,
        dx,
        dy,
        dxi,
        dpx,
        dpy,
        dpz,
        k_x,
        k_y,
        k_xi,
        k_px,
        k_py,
        k_pz,
    ) = bunch.get_rk4_arrays()
    ex, ey, ez, bx, by, bz = bunch.get_field_arrays()

    # Particle species constant (currently assumes electrons).
    q_over_mc = bunch.q_species / (bunch.m_species * ct.c)

    # Calculate push.
    for i in range(4):
        t_i = t

        # Calculate factors.
        if i in [0, 3]:
            fac1 = 1.0
            fac2 = 1.0 / 6.0
            if i == 3:
                t_i += dt
        else:
            fac1 = 0.5
            fac2 = 2.0 / 6.0
            t_i += dt / 2

        # Calculate contributions the push.
        if i == 0:
            # Gather field at the initial location of the particles.
            gather_fields(
                fields,
                bunch.x,
                bunch.y,
                bunch.xi,
                t_i,
                ex,
                ey,
                ez,
                bx,
                by,
                bz,
                bunch.name,
            )

            # Calculate k_1.
            calculate_k(
                k_x,
                k_y,
                k_xi,
                k_px,
                k_py,
                k_pz,
                q_over_mc,
                bunch.px,
                bunch.py,
                bunch.pz,
                ex,
                ey,
                ez,
                bx,
                by,
                bz,
            )

            # Initialize push with the contribution from k1.
            initialize_push(dx, k_x, fac2)
            initialize_push(dy, k_y, fac2)
            initialize_push(dxi, k_xi, fac2)
            initialize_push(dpx, k_px, fac2)
            initialize_push(dpy, k_py, fac2)
            initialize_push(dpz, k_pz, fac2)
        else:
            # Update particle coordinates to x = x_n + k_i * fac1
            update_coord(x, bunch.x, dt, k_x, fac1)
            update_coord(y, bunch.y, dt, k_y, fac1)
            update_coord(xi, bunch.xi, dt, k_xi, fac1)
            update_coord(px, bunch.px, dt, k_px, fac1)
            update_coord(py, bunch.py, dt, k_py, fac1)
            update_coord(pz, bunch.pz, dt, k_pz, fac1)

            # Gather field at updated positions.
            gather_fields(fields, x, y, xi, t_i, ex, ey, ez, bx, by, bz, bunch.name)

            # Calculate k_i.
            calculate_k(
                k_x,
                k_y,
                k_xi,
                k_px,
                k_py,
                k_pz,
                q_over_mc,
                px,
                py,
                pz,
                ex,
                ey,
                ez,
                bx,
                by,
                bz,
            )

            # Add the contribution of k_i to the push.
            update_push(dx, k_x, fac2)
            update_push(dy, k_y, fac2)
            update_push(dxi, k_xi, fac2)
            update_push(dpx, k_px, fac2)
            update_push(dpy, k_py, fac2)
            update_push(dpz, k_pz, fac2)

    # Apply push.
    apply_push(bunch.x, dt, dx)
    apply_push(bunch.y, dt, dy)
    apply_push(bunch.xi, dt, dxi)
    apply_push(bunch.px, dt, dpx)
    apply_push(bunch.py, dt, dpy)
    apply_push(bunch.pz, dt, dpz)


@njit_parallel()
def initialize_coord(x, x_0):
    for i in prange(x.shape[0]):
        x[i] = x_0[i]


@njit_parallel()
def update_coord(x, x_0, dt, k_x, fac):
    for i in prange(x.shape[0]):
        x[i] = x_0[i] + dt * k_x[i] * fac


@njit_parallel()
def initialize_push(dx, k_x, fac):
    for i in prange(dx.shape[0]):
        dx[i] = k_x[i] * fac


@njit_parallel()
def update_push(dx, k_x, fac):
    for i in prange(dx.shape[0]):
        dx[i] += k_x[i] * fac


@njit_parallel()
def apply_push(x, dt, dx):
    for i in prange(x.shape[0]):
        x[i] += dt * dx[i]


@njit_parallel(fastmath=True, error_model="numpy")
def calculate_k(
    k_x, k_y, k_xi, k_px, k_py, k_pz, q_over_mc, px, py, pz, ex, ey, ez, bx, by, bz
):
    for i in prange(k_x.shape[0]):
        px_i = px[i]
        py_i = py[i]
        pz_i = pz[i]
        ex_i = ex[i]
        ey_i = ey[i]
        ez_i = ez[i]
        bx_i = bx[i]
        by_i = by[i]
        bz_i = bz[i]

        c_over_gamma_i = ct.c / math.sqrt(1 + px_i**2 + py_i**2 + pz_i**2)
        vx_i = px_i * c_over_gamma_i
        vy_i = py_i * c_over_gamma_i
        vz_i = pz_i * c_over_gamma_i

        k_x[i] = vx_i
        k_y[i] = vy_i
        k_xi[i] = vz_i - ct.c
        k_px[i] = q_over_mc * (ex_i + vy_i * bz_i - vz_i * by_i)
        k_py[i] = q_over_mc * (ey_i - vx_i * bz_i + vz_i * bx_i)
        k_pz[i] = q_over_mc * (ez_i + vx_i * by_i - vy_i * bx_i)



"""Contains the logic for pushing a particle beam with transport matrices."""

import numpy as np

from wake_t.physics_models.beam_optics.transfer_matrices import (
    first_order_matrix,
    second_order_matrix,
)


def track_with_transfer_map(beam_matrix, z, L, theta, k1, k2, gamma_ref, order=2):
    """
    Track beam distribution throwgh beamline element by using a transfer map.
    This function was initially based on code from Ocelot (see
    https://github.com/ocelot-collab/ocelot) written by S. Tomin.

    Parameters
    ----------
    beam_matrix : array
        6 x N matrix, where N is the number of particles, containing the
        phase-space information of the bunch as (x, x', y, y', xi, dp) in
        units of (m, rad, m, rad, m, -). dp is defined as
        dp = (g-g_ref)/g_ref, while x' = px/p_kin and y' = py/p_kin, where
        p_kin is the kinetic momentum of each particle.
    z : float
        Longitudinal position in which to obtain the bunch distribution
    L : float
        Total length of the beamline element
    theta : float
        Bending angle of the beamline element
    k1 : float
        Quadrupole gradient of the beamline element in units of 1/m^2. A
        positive value implies focusing on the 'x' plane, while a negative
        gradient corresponds to focusing on 'y'
    k2 : float
        Sextupole gradient of the beamline element in units of 1/m^3. A
        positive value implies focusing on the 'x' plane, while a negative
        gradient corresponds to focusing on 'y'
    gamma_ref : float
        Reference energy with respect to which the particle momentum dp is
        calculated.
    order : int
        Indicates the order of the transport map to apply. Tracking up to
        second order is possible.

    """
    R = first_order_matrix(z, L, theta, k1, gamma_ref)
    bm_new = np.dot(R, beam_matrix)
    if order == 2:
        T = second_order_matrix(z, L, theta, k1, k2, gamma_ref)
        bm_new += np.einsum("ijk,j...,k...", T, beam_matrix, beam_matrix).T
    return bm_new



from typing import Optional, Callable

import numpy as np
import scipy.constants as ct

from .solver import calculate_wakefields
from wake_t.fields.rz_wakefield import RZWakefield
from wake_t.physics_models.laser.laser_pulse import LaserPulse


class Quasistatic2DWakefield(RZWakefield):
    """
    This class calculates the plasma wakefields using the gridless
    quasi-static model in r-z geometry originally developed by P. Baxevanis
    and G. Stupakov [1]_.

    The model implemented here includes additional features with respect to the
    original version in [1]_. Among them is the support for laser drivers,
    particle beams (instead of an analytic charge distribution), non-uniform
    and finite plasma density profiles, as well as an Adams-Bashforth pusher
    for the plasma particles (in addition the original Runke-Kutta pusher).

    As a kinetic quasi-static model in r-z geometry, it computes the plasma
    response by evolving a 1D radial plasma column from the front to the back
    of the simulation domain. A special feature of this model is that it does
    not need a grid in order to compute this evolution, and it allows the
    fields ``E`` and ``B`` to be calculated at any radial position in the
    plasma column.

    In the Wake-T implementation, a grid is used only in order to be able
    to easily interpolate the fields to the beam particles. After evolving the
    plasma column, ``E`` and ``B`` are calculated at the locations of the grid
    nodes. Similarly, the charge density ``rho`` and susceptibility ``chi``
    of the plasma are computed by depositing the charge of the plasma
    particles on the same grid. This useful for diagnostics and for evolving
    a laser pulse.

    Parameters
    ----------
    density_function : callable
        Function that returns the density value at the given position z.
        This parameter is given by the ``PlasmaStage`` and does not need
        to be specified by the user.
    r_max : float
        Maximum radial position up to which plasma wakefield will be
        calculated.
    xi_min : float
        Minimum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    xi_max : float
        Maximum longitudinal (speed of light frame) position up to which
        plasma wakefield will be calculated.
    n_r : int
        Number of grid elements along `r` to calculate the wakefields.
    n_xi : int
        Number of grid elements along `xi` to calculate the wakefields.
    ppc : int, optional
        Number of plasma particles per radial cell. By default ``ppc=2``.
    dz_fields : float, optional
        Determines how often the plasma wakefields should be updated.
        For example, if ``dz_fields=10e-6``, the plasma wakefields are
        only updated every time the simulation window advances by
        10 micron. By default ``dz_fields=xi_max-xi_min``, i.e., the
        length the simulation box.
    r_max_plasma : float, optional
        Maximum radial extension of the plasma column. If ``None``, the
        plasma extends up to the ``r_max`` boundary of the simulation box.
    parabolic_coefficient : float or callable, optional
        The coefficient for the transverse parabolic density profile. The
        radial density distribution is calculated as
        ``n_r = n_p * (1 + parabolic_coefficient * r**2)``, where n_p is
        the local on-axis plasma density. If a ``float`` is provided, the
        same value will be used throwout the stage. Alternatively, a
        function which returns the value of the coefficient at the given
        position ``z`` (e.g. ``def func(z)``) might also be provided.
    p_shape : str, optional
        Particle shape to be used for the beam charge deposition. Possible
        values are ``'linear'`` or ``'cubic'`` (default).
    max_gamma : float, optional
        Plasma particles whose ``gamma`` exceeds ``max_gamma`` are
        considered to violate the quasistatic condition and are put at
        rest (i.e., ``gamma=1.``, ``pr=pz=0.``). By default
        ``max_gamma=10``.
    plasma_pusher : str, optional
        The pusher used to evolve the plasma particles. Possible values
        are ``'rk4'`` (Runge-Kutta 4th order) or ``'ab5'`` (Adams-Bashforth
        5th order).
    laser : LaserPulse, optional
        Laser driver of the plasma stage.
    laser_evolution : bool, optional
        If True (default), the laser pulse is evolved
        using a laser envelope model. If ``False``, the pulse envelope
        stays unchanged throughout the computation.
    laser_envelope_substeps : int, optional
        Number of substeps of the laser envelope solver per ``dz_fields``.
        The time step of the envelope solver is therefore
        ``dz_fields / c / laser_envelope_substeps``.
    laser_envelope_nxi, laser_envelope_nr : int, optional
        If given, the laser envelope will run in a grid of size
        (``laser_envelope_nxi``, ``laser_envelope_nr``) instead
        of (``n_xi``, ``n_r``). This allows the laser to run in a finer (or
        coarser) grid than the plasma wake. It is not necessary to specify
        both parameters. If one of them is not given, the resolution of
        the plasma grid with be used for that direction.
    laser_envelope_use_phase : bool, optional
        Determines whether to take into account the terms related to the
        longitudinal derivative of the complex phase in the envelope
        solver.

    References
    ----------
    .. [1] P. Baxevanis and G. Stupakov, "Novel fast simulation technique
        for axisymmetric plasma wakefield acceleration configurations in
        the blowout regime," Phys. Rev. Accel. Beams 21, 071301 (2018),
        https://link.aps.org/doi/10.1103/PhysRevAccelBeams.21.071301

    """

    def __init__(
        self,
        density_function: Callable[[float], float],
        r_max: float,
        xi_min: float,
        xi_max: float,
        n_r: int,
        n_xi: int,
        ppc: Optional[int] = 2,
        dz_fields: Optional[float] = None,
        r_max_plasma: Optional[float] = None,
        parabolic_coefficient: Optional[float] = 0.0,
        p_shape: Optional[str] = "cubic",
        max_gamma: Optional[float] = 10,
        plasma_pusher: Optional[str] = "rk4",
        laser: Optional[LaserPulse] = None,
        laser_evolution: Optional[bool] = True,
        laser_envelope_substeps: Optional[int] = 1,
        laser_envelope_nxi: Optional[int] = None,
        laser_envelope_nr: Optional[int] = None,
        laser_envelope_use_phase: Optional[bool] = True,
    ) -> None:
        self.ppc = ppc
        self.r_max_plasma = r_max_plasma
        self.parabolic_coefficient = self._get_parabolic_coefficient_fn(
            parabolic_coefficient
        )
        self.p_shape = p_shape
        self.max_gamma = max_gamma
        self.plasma_pusher = plasma_pusher
        super().__init__(
            density_function=density_function,
            r_max=r_max,
            xi_min=xi_min,
            xi_max=xi_max,
            n_r=n_r,
            n_xi=n_xi,
            dz_fields=dz_fields,
            laser=laser,
            laser_evolution=laser_evolution,
            laser_envelope_substeps=laser_envelope_substeps,
            laser_envelope_nxi=laser_envelope_nxi,
            laser_envelope_nr=laser_envelope_nr,
            laser_envelope_use_phase=laser_envelope_use_phase,
            model_name="quasistatic_2d",
        )

    def _calculate_wakefield(self, bunches):
        parabolic_coefficient = self.parabolic_coefficient(self.t * ct.c)

        # Get square of laser envelope
        if self.laser is not None:
            a_env_2 = np.abs(self.laser.get_envelope()) ** 2
            # If linearly polarized, divide by 2 so that the ponderomotive
            # force on the plasma particles is correct.
            if self.laser.polarization == "linear":
                a_env_2 /= 2
        else:
            a_env_2 = np.zeros((self.n_xi, self.n_r))

        # Calculate plasma wakefields
        calculate_wakefields(
            a_env_2,
            bunches,
            self.r_max,
            self.xi_min,
            self.xi_max,
            self.n_r,
            self.n_xi,
            self.ppc,
            self.n_p,
            r_max_plasma=self.r_max_plasma,
            parabolic_coefficient=parabolic_coefficient,
            p_shape=self.p_shape,
            max_gamma=self.max_gamma,
            plasma_pusher=self.plasma_pusher,
            fld_arrays=[
                self.rho,
                self.chi,
                self.e_r,
                self.e_z,
                self.b_t,
                self.xi_fld,
                self.r_fld,
            ],
        )

    def _get_parabolic_coefficient_fn(self, parabolic_coefficient):
        """Get parabolic_coefficient profile function"""
        if isinstance(parabolic_coefficient, float):

            def uniform_parabolic_coefficient(z):
                return np.ones_like(z) * parabolic_coefficient

            return uniform_parabolic_coefficient
        elif callable(parabolic_coefficient):
            return parabolic_coefficient
        else:
            raise ValueError(
                "Type {} not supported for parabolic_coefficient.".format(
                    type(parabolic_coefficient)
                )
            )



from .tm_elements import Drift, Dipole, Quadrupole, Sextupole, TMElement
from .plasma_stage import PlasmaStage
from .plasma_ramp import PlasmaRamp
from .active_plasma_lens import ActivePlasmaLens
from .beamline import Beamline
from .field_element import FieldElement
from .field_quadrupole import FieldQuadrupole


__all__ = [
    "Drift",
    "Dipole",
    "Quadrupole",
    "Sextupole",
    "PlasmaStage",
    "PlasmaRamp",
    "ActivePlasmaLens",
    "Beamline",
    "FieldElement",
    "TMElement",
    "FieldQuadrupole",
]
